#    Copyright 2015 SimpliVity Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt
from oslo_utils import timeutils
import pytz
import six

from cinder import context
from cinder import exception
from cinder import objects
from cinder.objects import base as ovo_base
from cinder.objects import fields
from cinder.tests.unit.consistencygroup import fake_consistencygroup
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit import objects as test_objects


@ddt.ddt
class TestVolume(test_objects.BaseObjectsTestCase):
    @staticmethod
    def _compare(test, db, obj):
        db = {k: v for k, v in db.items()
              if not k.endswith('metadata') or k.startswith('volume')}
        test_objects.BaseObjectsTestCase._compare(test, db, obj)

    @mock.patch('cinder.db.sqlalchemy.api.volume_get')
    def test_get_by_id(self, volume_get):
        db_volume = fake_volume.fake_db_volume()
        volume_get.return_value = db_volume
        volume = objects.Volume.get_by_id(self.context, fake.VOLUME_ID)
        volume_get.assert_called_once_with(self.context, fake.VOLUME_ID)
        self._compare(self, db_volume, volume)

    @mock.patch('cinder.db.sqlalchemy.api.model_query')
    def test_get_by_id_no_existing_id(self, model_query):
        pf = (model_query().options().options().options().options().options().
              options())
        pf.filter_by().first.return_value = None
        self.assertRaises(exception.VolumeNotFound,
                          objects.Volume.get_by_id, self.context, 123)

    @mock.patch('cinder.db.volume_create')
    def test_create(self, volume_create):
        db_volume = fake_volume.fake_db_volume()
        volume_create.return_value = db_volume
        volume = objects.Volume(context=self.context)
        volume.create()
        self.assertEqual(db_volume['id'], volume.id)

    @mock.patch('cinder.db.volume_update')
    @ddt.data(False, True)
    def test_save(self, test_cg, volume_update):
        db_volume = fake_volume.fake_db_volume()
        volume = objects.Volume._from_db_object(self.context,
                                                objects.Volume(), db_volume)
        volume.display_name = 'foobar'
        if test_cg:
            volume.consistencygroup = None
        volume.save()
        volume_update.assert_called_once_with(self.context, volume.id,
                                              {'display_name': 'foobar'})

    def test_save_error(self):
        db_volume = fake_volume.fake_db_volume()
        volume = objects.Volume._from_db_object(self.context,
                                                objects.Volume(), db_volume)
        volume.display_name = 'foobar'
        volume.consistencygroup = (
            fake_consistencygroup.fake_consistencyobject_obj(self.context))
        self.assertRaises(exception.ObjectActionError,
                          volume.save)

    @mock.patch('cinder.db.volume_metadata_update',
                return_value={'key1': 'value1'})
    @mock.patch('cinder.db.volume_update')
    def test_save_with_metadata(self, volume_update, metadata_update):
        db_volume = fake_volume.fake_db_volume()
        volume = objects.Volume._from_db_object(self.context,
                                                objects.Volume(), db_volume)
        volume.display_name = 'foobar'
        volume.metadata = {'key1': 'value1'}
        self.assertEqual({'display_name': 'foobar',
                          'metadata': {'key1': 'value1'}},
                         volume.obj_get_changes())
        volume.save()
        volume_update.assert_called_once_with(self.context, volume.id,
                                              {'display_name': 'foobar'})
        metadata_update.assert_called_once_with(self.context, volume.id,
                                                {'key1': 'value1'}, True)

    @mock.patch('cinder.db.volume_admin_metadata_update',
                return_value={'key1': 'value1'})
    @mock.patch('cinder.db.volume_update')
    def test_save_with_admin_metadata(self, volume_update,
                                      admin_metadata_update):
        # Test with no admin context
        db_volume = fake_volume.fake_db_volume()
        volume = objects.Volume._from_db_object(self.context,
                                                objects.Volume(), db_volume)
        volume.admin_metadata = {'key1': 'value1'}
        volume.save()
        self.assertFalse(admin_metadata_update.called)

        # Test with admin context
        admin_context = context.RequestContext(self.user_id, self.project_id,
                                               is_admin=True)
        volume = objects.Volume._from_db_object(admin_context,
                                                objects.Volume(), db_volume)
        volume.admin_metadata = {'key1': 'value1'}
        volume.save()
        admin_metadata_update.assert_called_once_with(
            admin_context, volume.id, {'key1': 'value1'}, True)

    def test_save_with_glance_metadata(self):
        db_volume = fake_volume.fake_db_volume()
        volume = objects.Volume._from_db_object(self.context,
                                                objects.Volume(), db_volume)
        volume.display_name = 'foobar'
        volume.glance_metadata = {'key1': 'value1'}
        self.assertRaises(exception.ObjectActionError, volume.save)

    def test_save_with_consistencygroup(self):
        db_volume = fake_volume.fake_db_volume()
        volume = objects.Volume._from_db_object(self.context,
                                                objects.Volume(), db_volume)
        volume.display_name = 'foobar'
        volume.consistencygroup = objects.ConsistencyGroup()
        self.assertRaises(exception.ObjectActionError, volume.save)

    def test_save_with_snapshots(self):
        db_volume = fake_volume.fake_db_volume()
        volume = objects.Volume._from_db_object(self.context,
                                                objects.Volume(), db_volume)
        volume.display_name = 'foobar'
        volume.snapshots = objects.SnapshotList()
        self.assertRaises(exception.ObjectActionError, volume.save)

    @mock.patch('oslo_utils.timeutils.utcnow', return_value=timeutils.utcnow())
    @mock.patch('cinder.db.sqlalchemy.api.volume_destroy')
    def test_destroy(self, volume_destroy, utcnow_mock):
        volume_destroy.return_value = {
            'status': 'deleted',
            'deleted': True,
            'deleted_at': utcnow_mock.return_value}
        db_volume = fake_volume.fake_db_volume()
        volume = objects.Volume._from_db_object(self.context,
                                                objects.Volume(), db_volume)
        volume.destroy()
        self.assertTrue(volume_destroy.called)
        admin_context = volume_destroy.call_args[0][0]
        self.assertTrue(admin_context.is_admin)
        self.assertTrue(volume.deleted)
        self.assertEqual('deleted', volume.status)
        self.assertEqual(utcnow_mock.return_value.replace(tzinfo=pytz.UTC),
                         volume.deleted_at)
        self.assertIsNone(volume.migration_status)

    def test_obj_fields(self):
        volume = objects.Volume(context=self.context, id=fake.VOLUME_ID,
                                name_id=fake.VOLUME_NAME_ID)
        self.assertEqual(['name', 'name_id', 'volume_metadata',
                          'volume_admin_metadata', 'volume_glance_metadata'],
                         volume.obj_extra_fields)
        self.assertEqual('volume-%s' % fake.VOLUME_NAME_ID, volume.name)
        self.assertEqual(fake.VOLUME_NAME_ID, volume.name_id)

    def test_obj_field_previous_status(self):
        volume = objects.Volume(context=self.context,
                                previous_status='backing-up')
        self.assertEqual('backing-up', volume.previous_status)

    @mock.patch('cinder.db.volume_metadata_delete')
    def test_delete_metadata_key(self, metadata_delete):
        volume = objects.Volume(self.context, id=fake.VOLUME_ID)
        volume.metadata = {'key1': 'value1', 'key2': 'value2'}
        self.assertEqual({}, volume._orig_metadata)
        volume.delete_metadata_key('key2')
        self.assertEqual({'key1': 'value1'}, volume.metadata)
        metadata_delete.assert_called_once_with(self.context, fake.VOLUME_ID,
                                                'key2')

    @mock.patch('cinder.db.volume_metadata_get')
    @mock.patch('cinder.db.volume_glance_metadata_get')
    @mock.patch('cinder.db.volume_admin_metadata_get')
    @mock.patch('cinder.objects.volume_type.VolumeType.get_by_id')
    @mock.patch('cinder.objects.volume_attachment.VolumeAttachmentList.'
                'get_all_by_volume_id')
    @mock.patch('cinder.objects.consistencygroup.ConsistencyGroup.get_by_id')
    @mock.patch('cinder.objects.snapshot.SnapshotList.get_all_for_volume')
    def test_obj_load_attr(self, mock_sl_get_all_for_volume, mock_cg_get_by_id,
                           mock_va_get_all_by_vol, mock_vt_get_by_id,
                           mock_admin_metadata_get, mock_glance_metadata_get,
                           mock_metadata_get):
        fake_db_volume = fake_volume.fake_db_volume(
            consistencygroup_id=fake.CONSISTENCY_GROUP_ID)
        volume = objects.Volume._from_db_object(
            self.context, objects.Volume(), fake_db_volume)

        # Test metadata lazy-loaded field
        metadata = {'foo': 'bar'}
        mock_metadata_get.return_value = metadata
        self.assertEqual(metadata, volume.metadata)
        mock_metadata_get.assert_called_once_with(self.context, volume.id)

        # Test glance_metadata lazy-loaded field
        glance_metadata = [{'key': 'foo', 'value': 'bar'}]
        mock_glance_metadata_get.return_value = glance_metadata
        self.assertEqual({'foo': 'bar'}, volume.glance_metadata)
        mock_glance_metadata_get.assert_called_once_with(
            self.context, volume.id)

        # Test volume_type lazy-loaded field
        # Case1. volume.volume_type_id = None
        self.assertIsNone(volume.volume_type)

        # Case2. volume2.volume_type_id = 1
        fake2 = fake_volume.fake_db_volume()
        fake2.update({'volume_type_id': fake.VOLUME_ID})
        volume2 = objects.Volume._from_db_object(
            self.context, objects.Volume(), fake2)
        volume_type = objects.VolumeType(context=self.context,
                                         id=fake.VOLUME_TYPE_ID)
        mock_vt_get_by_id.return_value = volume_type
        self.assertEqual(volume_type, volume2.volume_type)
        mock_vt_get_by_id.assert_called_once_with(self.context,
                                                  volume2.volume_type_id)

        # Test consistencygroup lazy-loaded field
        consistencygroup = objects.ConsistencyGroup(
            context=self.context, id=fake.CONSISTENCY_GROUP_ID)
        mock_cg_get_by_id.return_value = consistencygroup
        self.assertEqual(consistencygroup, volume.consistencygroup)
        mock_cg_get_by_id.assert_called_once_with(self.context,
                                                  volume.consistencygroup_id)

        # Test snapshots lazy-loaded field
        snapshots = objects.SnapshotList(context=self.context,
                                         id=fake.SNAPSHOT_ID)
        mock_sl_get_all_for_volume.return_value = snapshots
        self.assertEqual(snapshots, volume.snapshots)
        mock_sl_get_all_for_volume.assert_called_once_with(self.context,
                                                           volume.id)

        # Test volume_attachment lazy-loaded field
        va_objs = [objects.VolumeAttachment(context=self.context, id=i)
                   for i in [fake.OBJECT_ID, fake.OBJECT2_ID, fake.OBJECT3_ID]]
        va_list = objects.VolumeAttachmentList(context=self.context,
                                               objects=va_objs)
        mock_va_get_all_by_vol.return_value = va_list
        self.assertEqual(va_list, volume.volume_attachment)
        mock_va_get_all_by_vol.assert_called_once_with(self.context, volume.id)

        # Test admin_metadata lazy-loaded field - user context
        adm_metadata = {'bar': 'foo'}
        mock_admin_metadata_get.return_value = adm_metadata
        self.assertEqual({}, volume.admin_metadata)
        self.assertFalse(mock_admin_metadata_get.called)

        # Test admin_metadata lazy-loaded field - admin context
        adm_context = self.context.elevated()
        volume = objects.Volume._from_db_object(adm_context, objects.Volume(),
                                                fake_volume.fake_db_volume())
        adm_metadata = {'bar': 'foo'}
        mock_admin_metadata_get.return_value = adm_metadata
        self.assertEqual(adm_metadata, volume.admin_metadata)
        mock_admin_metadata_get.assert_called_once_with(adm_context, volume.id)

    @mock.patch('cinder.objects.consistencygroup.ConsistencyGroup.get_by_id')
    def test_obj_load_attr_cgroup_not_exist(self, mock_cg_get_by_id):
        fake_db_volume = fake_volume.fake_db_volume(consistencygroup_id=None)
        volume = objects.Volume._from_db_object(
            self.context, objects.Volume(), fake_db_volume)

        self.assertIsNone(volume.consistencygroup)
        mock_cg_get_by_id.assert_not_called()

    @mock.patch('cinder.objects.group.Group.get_by_id')
    def test_obj_load_attr_group_not_exist(self, mock_group_get_by_id):
        fake_db_volume = fake_volume.fake_db_volume(group_id=None)
        volume = objects.Volume._from_db_object(
            self.context, objects.Volume(), fake_db_volume)

        self.assertIsNone(volume.group)
        mock_group_get_by_id.assert_not_called()

    def test_from_db_object_with_all_expected_attributes(self):
        expected_attrs = ['metadata', 'admin_metadata', 'glance_metadata',
                          'volume_type', 'volume_attachment',
                          'consistencygroup']

        db_metadata = [{'key': 'foo', 'value': 'bar'}]
        db_admin_metadata = [{'key': 'admin_foo', 'value': 'admin_bar'}]
        db_glance_metadata = [{'key': 'glance_foo', 'value': 'glance_bar'}]
        db_volume_type = fake_volume.fake_db_volume_type()
        db_volume_attachments = fake_volume.volume_attachment_db_obj()
        db_consistencygroup = fake_consistencygroup.fake_db_consistencygroup()
        db_snapshots = fake_snapshot.fake_db_snapshot()

        db_volume = fake_volume.fake_db_volume(
            volume_metadata=db_metadata,
            volume_admin_metadata=db_admin_metadata,
            volume_glance_metadata=db_glance_metadata,
            volume_type=db_volume_type,
            volume_attachment=[db_volume_attachments],
            consistencygroup=db_consistencygroup,
            snapshots=[db_snapshots],
        )
        volume = objects.Volume._from_db_object(self.context, objects.Volume(),
                                                db_volume, expected_attrs)

        self.assertEqual({'foo': 'bar'}, volume.metadata)
        self.assertEqual({'admin_foo': 'admin_bar'}, volume.admin_metadata)
        self.assertEqual({'glance_foo': 'glance_bar'}, volume.glance_metadata)
        self._compare(self, db_volume_type, volume.volume_type)
        self._compare(self, db_volume_attachments, volume.volume_attachment)
        self._compare(self, db_consistencygroup, volume.consistencygroup)
        self._compare(self, db_snapshots, volume.snapshots)

    @mock.patch('cinder.db.volume_glance_metadata_get', return_value={})
    @mock.patch('cinder.db.sqlalchemy.api.volume_get')
    def test_refresh(self, volume_get, volume_metadata_get):
        db_volume1 = fake_volume.fake_db_volume()
        db_volume2 = db_volume1.copy()
        db_volume2['display_name'] = 'foobar'

        # On the second volume_get, return the volume with an updated
        # display_name
        volume_get.side_effect = [db_volume1, db_volume2]
        volume = objects.Volume.get_by_id(self.context, fake.VOLUME_ID)
        self._compare(self, db_volume1, volume)

        # display_name was updated, so a volume refresh should have a new value
        # for that field
        volume.refresh()
        self._compare(self, db_volume2, volume)
        if six.PY3:
            call_bool = mock.call.__bool__()
        else:
            call_bool = mock.call.__nonzero__()
        volume_get.assert_has_calls([mock.call(self.context, fake.VOLUME_ID),
                                     call_bool,
                                     mock.call(self.context, fake.VOLUME_ID)])

    def test_metadata_aliases(self):
        volume = objects.Volume(context=self.context)
        # metadata<->volume_metadata
        volume.metadata = {'abc': 'def'}
        self.assertEqual([{'key': 'abc', 'value': 'def'}],
                         volume.volume_metadata)

        md = [{'key': 'def', 'value': 'abc'}]
        volume.volume_metadata = md
        self.assertEqual({'def': 'abc'}, volume.metadata)

        # admin_metadata<->volume_admin_metadata
        volume.admin_metadata = {'foo': 'bar'}
        self.assertEqual([{'key': 'foo', 'value': 'bar'}],
                         volume.volume_admin_metadata)

        volume.volume_admin_metadata = [{'key': 'xyz', 'value': '42'}]
        self.assertEqual({'xyz': '42'}, volume.admin_metadata)

        # glance_metadata<->volume_glance_metadata
        volume.glance_metadata = {'jkl': 'mno'}
        self.assertEqual([{'key': 'jkl', 'value': 'mno'}],
                         volume.volume_glance_metadata)

        volume.volume_glance_metadata = [{'key': 'prs', 'value': 'tuw'}]
        self.assertEqual({'prs': 'tuw'}, volume.glance_metadata)

    @mock.patch('cinder.db.volume_metadata_update', return_value={})
    @mock.patch('cinder.db.volume_update')
    @ddt.data({'src_vol_type_id': fake.VOLUME_TYPE_ID,
               'dest_vol_type_id': fake.VOLUME_TYPE2_ID},
              {'src_vol_type_id': None,
               'dest_vol_type_id': fake.VOLUME_TYPE2_ID})
    @ddt.unpack
    def test_finish_volume_migration(self, volume_update, metadata_update,
                                     src_vol_type_id, dest_vol_type_id):
        src_volume_db = fake_volume.fake_db_volume(
            **{'id': fake.VOLUME_ID, 'volume_type_id': src_vol_type_id})
        if src_vol_type_id:
            src_volume_db['volume_type'] = fake_volume.fake_db_volume_type(
                id=src_vol_type_id)
        dest_volume_db = fake_volume.fake_db_volume(
            **{'id': fake.VOLUME2_ID, 'volume_type_id': dest_vol_type_id})
        if dest_vol_type_id:
            dest_volume_db['volume_type'] = fake_volume.fake_db_volume_type(
                id=dest_vol_type_id)
        expected_attrs = objects.Volume._get_expected_attrs(self.context)
        src_volume = objects.Volume._from_db_object(
            self.context, objects.Volume(), src_volume_db,
            expected_attrs=expected_attrs)
        dest_volume = objects.Volume._from_db_object(
            self.context, objects.Volume(), dest_volume_db,
            expected_attrs=expected_attrs)
        updated_dest_volume = src_volume.finish_volume_migration(
            dest_volume)
        self.assertEqual('deleting', updated_dest_volume.migration_status)
        self.assertEqual('migration src for ' + src_volume.id,
                         updated_dest_volume.display_description)
        self.assertEqual(src_volume.id, updated_dest_volume._name_id)
        self.assertTrue(volume_update.called)
        volume_update.assert_has_calls([
            mock.call(self.context, src_volume.id, mock.ANY),
            mock.call(self.context, dest_volume.id, mock.ANY)])
        ctxt, vol_id, updates = volume_update.call_args[0]
        self.assertNotIn('volume_type', updates)

        # Ensure that the destination volume type has not been overwritten
        self.assertEqual(dest_vol_type_id,
                         getattr(updated_dest_volume, 'volume_type_id'))
        # Ignore these attributes, since they were updated by
        # finish_volume_migration
        ignore_keys = ('id', 'provider_location', '_name_id',
                       'migration_status', 'display_description', 'status',
                       'volume_glance_metadata', 'volume_type')

        dest_vol_dict = {k: updated_dest_volume[k] for k in
                         updated_dest_volume.keys() if k not in ignore_keys}
        src_vol_dict = {k: src_volume[k] for k in src_volume.keys()
                        if k not in ignore_keys}
        self.assertEqual(src_vol_dict, dest_vol_dict)

    def test_volume_with_metadata_serialize_deserialize_no_changes(self):
        updates = {'volume_glance_metadata': [{'key': 'foo', 'value': 'bar'}],
                   'expected_attrs': ['glance_metadata']}
        volume = fake_volume.fake_volume_obj(self.context, **updates)
        serializer = objects.base.CinderObjectSerializer()
        serialized_volume = serializer.serialize_entity(self.context, volume)
        volume = serializer.deserialize_entity(self.context, serialized_volume)
        self.assertDictEqual({}, volume.obj_get_changes())

    @mock.patch('cinder.db.volume_admin_metadata_update')
    @mock.patch('cinder.db.sqlalchemy.api.volume_attach')
    def test_begin_attach(self, volume_attach, metadata_update):
        volume = fake_volume.fake_volume_obj(self.context)
        db_attachment = fake_volume.volume_attachment_db_obj(
            volume_id=volume.id,
            attach_status=fields.VolumeAttachStatus.ATTACHING)
        volume_attach.return_value = db_attachment
        metadata_update.return_value = {'attached_mode': 'rw'}

        with mock.patch.object(self.context, 'elevated') as mock_elevated:
            mock_elevated.return_value = context.get_admin_context()
            attachment = volume.begin_attach("rw")
            self.assertIsInstance(attachment, objects.VolumeAttachment)
            self.assertEqual(volume.id, attachment.volume_id)
            self.assertEqual(fields.VolumeAttachStatus.ATTACHING,
                             attachment.attach_status)
            metadata_update.assert_called_once_with(self.context.elevated(),
                                                    volume.id,
                                                    {'attached_mode': u'rw'},
                                                    True)
            self.assertEqual('rw', volume.admin_metadata['attached_mode'])

    @mock.patch('cinder.db.volume_admin_metadata_delete')
    @mock.patch('cinder.db.sqlalchemy.api.volume_detached')
    @mock.patch('cinder.objects.volume_attachment.VolumeAttachmentList.'
                'get_all_by_volume_id')
    def test_volume_detached_with_attachment(
            self, volume_attachment_get,
            volume_detached,
            metadata_delete):
        va_objs = [objects.VolumeAttachment(context=self.context, id=i)
                   for i in [fake.OBJECT_ID, fake.OBJECT2_ID, fake.OBJECT3_ID]]
        # As changes are not saved, we need reset it here. Later changes
        # will be checked.
        for obj in va_objs:
            obj.obj_reset_changes()
        va_list = objects.VolumeAttachmentList(context=self.context,
                                               objects=va_objs)
        va_list.obj_reset_changes()
        volume_attachment_get.return_value = va_list
        admin_context = context.get_admin_context()
        volume = fake_volume.fake_volume_obj(
            admin_context,
            volume_attachment=va_list,
            volume_admin_metadata=[{'key': 'attached_mode',
                                    'value': 'rw'}])
        self.assertEqual(3, len(volume.volume_attachment))
        volume_detached.return_value = ({'status': 'in-use'},
                                        {'attached_mode': 'rw'})
        with mock.patch.object(admin_context, 'elevated') as mock_elevated:
            mock_elevated.return_value = admin_context
            volume.finish_detach(fake.OBJECT_ID)
            volume_detached.assert_called_once_with(admin_context,
                                                    volume.id,
                                                    fake.OBJECT_ID)
            metadata_delete.assert_called_once_with(admin_context,
                                                    volume.id,
                                                    'attached_mode')
            self.assertEqual('in-use', volume.status)
            self.assertEqual({}, volume.cinder_obj_get_changes())
            self.assertEqual(2, len(volume.volume_attachment))
            self.assertNotIn('attached_mode', volume.admin_metadata)

    @mock.patch('cinder.db.volume_admin_metadata_delete')
    @mock.patch('cinder.db.sqlalchemy.api.volume_detached')
    @mock.patch('cinder.objects.volume_attachment.VolumeAttachmentList.'
                'get_all_by_volume_id')
    def test_volume_detached_without_attachment(
            self, volume_attachment_get, volume_detached, metadata_delete):
        admin_context = context.get_admin_context()
        volume = fake_volume.fake_volume_obj(
            admin_context,
            volume_admin_metadata=[{'key': 'attached_mode',
                                    'value': 'rw'}])
        self.assertFalse(volume.obj_attr_is_set('volume_attachment'))
        volume_detached.return_value = ({'status': 'in-use'}, None)
        with mock.patch.object(admin_context, 'elevated') as mock_elevated:
            mock_elevated.return_value = admin_context
            volume.finish_detach(fake.OBJECT_ID)
            metadata_delete.assert_called_once_with(admin_context,
                                                    volume.id,
                                                    'attached_mode')
            volume_detached.assert_called_once_with(admin_context,
                                                    volume.id,
                                                    fake.OBJECT_ID)
            self.assertEqual('in-use', volume.status)
            self.assertEqual({}, volume.cinder_obj_get_changes())
            self.assertFalse(volume_attachment_get.called)

    @ddt.data('1.6', '1.7')
    def test_obj_make_compatible_cluster_added(self, version):
        extra_data = {'cluster_name': 'cluster_name',
                      'cluster': objects.Cluster()}
        volume = objects.Volume(self.context, host='host', **extra_data)

        serializer = ovo_base.CinderObjectSerializer(version)
        primitive = serializer.serialize_entity(self.context, volume)

        converted_volume = objects.Volume.obj_from_primitive(primitive)
        is_set = version == '1.7'
        for key in extra_data:
            self.assertEqual(is_set, converted_volume.obj_attr_is_set(key))
        self.assertEqual('host', converted_volume.host)

    @ddt.data('1.9', '1.10')
    def test_obj_make_compatible_groups_added(self, version):
        extra_data = {'group_id': fake.GROUP_ID,
                      'group': objects.Group()}
        volume = objects.Volume(self.context, host='host', **extra_data)

        serializer = ovo_base.CinderObjectSerializer(version)
        primitive = serializer.serialize_entity(self.context, volume)

        converted_volume = objects.Volume.obj_from_primitive(primitive)
        is_set = version == '1.10'
        for key in extra_data:
            self.assertEqual(is_set, converted_volume.obj_attr_is_set(key))
        self.assertEqual('host', converted_volume.host)

    @ddt.data(True, False)
    def test_is_replicated(self, result):
        volume_type = fake_volume.fake_volume_type_obj(self.context)
        volume = fake_volume.fake_volume_obj(
            self.context, volume_type_id=volume_type.id)
        volume.volume_type = volume_type
        with mock.patch.object(volume_type, 'is_replicated',
                               return_value=result) as is_replicated:
            self.assertEqual(result, volume.is_replicated())
            is_replicated.assert_called_once_with()

    def test_is_replicated_no_type(self):
        volume = fake_volume.fake_volume_obj(
            self.context, volume_type_id=None, volume_type=None)
        self.assertFalse(bool(volume.is_replicated()))


@ddt.ddt
class TestVolumeList(test_objects.BaseObjectsTestCase):
    @mock.patch('cinder.db.volume_get_all')
    def test_get_all(self, volume_get_all):
        db_volume = fake_volume.fake_db_volume()
        volume_get_all.return_value = [db_volume]

        volumes = objects.VolumeList.get_all(self.context,
                                             mock.sentinel.marker,
                                             mock.sentinel.limit,
                                             mock.sentinel.sort_key,
                                             mock.sentinel.sort_dir)
        self.assertEqual(1, len(volumes))
        TestVolume._compare(self, db_volume, volumes[0])

    @mock.patch('cinder.db.volume_get_all_by_host')
    def test_get_by_host(self, get_all_by_host):
        db_volume = fake_volume.fake_db_volume()
        get_all_by_host.return_value = [db_volume]

        volumes = objects.VolumeList.get_all_by_host(
            self.context, 'fake-host')
        self.assertEqual(1, len(volumes))
        TestVolume._compare(self, db_volume, volumes[0])

    @mock.patch('cinder.db.volume_get_all_by_group')
    def test_get_by_group(self, get_all_by_group):
        db_volume = fake_volume.fake_db_volume()
        get_all_by_group.return_value = [db_volume]

        volumes = objects.VolumeList.get_all_by_group(
            self.context, 'fake-host')
        self.assertEqual(1, len(volumes))
        TestVolume._compare(self, db_volume, volumes[0])

    @mock.patch('cinder.db.volume_get_all_by_project')
    def test_get_by_project(self, get_all_by_project):
        db_volume = fake_volume.fake_db_volume()
        get_all_by_project.return_value = [db_volume]

        volumes = objects.VolumeList.get_all_by_project(
            self.context, mock.sentinel.project_id, mock.sentinel.marker,
            mock.sentinel.limit, mock.sentinel.sorted_keys,
            mock.sentinel.sorted_dirs, mock.sentinel.filters)
        self.assertEqual(1, len(volumes))
        TestVolume._compare(self, db_volume, volumes[0])

    @ddt.data(['name_id'], ['__contains__'])
    def test_get_by_project_with_sort_key(self, sort_keys):
        fake_volume.fake_db_volume()

        self.assertRaises(exception.InvalidInput,
                          objects.VolumeList.get_all_by_project,
                          self.context,
                          self.context.project_id,
                          sort_keys=sort_keys)

    @mock.patch('cinder.db.volume_include_in_cluster')
    def test_include_in_cluster(self, include_mock):
        filters = {'host': mock.sentinel.host,
                   'cluster_name': mock.sentinel.cluster_name}
        cluster = 'new_cluster'
        objects.VolumeList.include_in_cluster(self.context, cluster, **filters)
        include_mock.assert_called_once_with(self.context, cluster, True,
                                             **filters)

    @mock.patch('cinder.db.volume_include_in_cluster')
    def test_include_in_cluster_specify_partial(self, include_mock):
        filters = {'host': mock.sentinel.host,
                   'cluster_name': mock.sentinel.cluster_name}
        cluster = 'new_cluster'
        objects.VolumeList.include_in_cluster(self.context, cluster,
                                              mock.sentinel.partial_rename,
                                              **filters)
        include_mock.assert_called_once_with(self.context, cluster,
                                             mock.sentinel.partial_rename,
                                             **filters)



# Copyright (C) 2017 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Schema for V3 Group types API.

"""

from cinder.api.validation import parameter_types


create = {
    'type': 'object',
    'properties': {
        'type': 'object',
        'group_type': {
            'type': 'object',
            'properties': {
                'name': parameter_types.name,
                'description': parameter_types.description,
                'is_public': parameter_types.boolean,
                'group_specs': parameter_types.metadata_allows_null,
            },
            'required': ['name'],
            'additionalProperties': False,
        },
    },
    'required': ['group_type'],
    'additionalProperties': False,
}

update = {
    'type': 'object',
    'properties': {
        'type': 'object',
        'group_type': {
            'type': 'object',
            'properties': {
                'name': parameter_types.name_allow_zero_min_length,
                'description': parameter_types.description,
                'is_public': parameter_types.boolean,
            },
            'additionalProperties': False,
        },
    },
    'required': ['group_type'],
    'additionalProperties': False,
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from cinder.common import sqlalchemyutils
from cinder import context
from cinder.db.sqlalchemy import api as db_api
from cinder.db.sqlalchemy import models
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import test


class TestPaginateQuery(test.TestCase):
    def setUp(self):
        super(TestPaginateQuery, self).setUp()
        self.ctxt = context.RequestContext(fake.USER_ID, fake.PROJECT_ID,
                                           auth_token=True,
                                           is_admin=True)
        self.query = db_api._volume_get_query(self.ctxt)
        self.model = models.Volume

    def test_paginate_query_marker_null(self):
        marker_object = self.model()
        self.assertIsNone(marker_object.display_name)
        self.assertIsNone(marker_object.updated_at)

        marker_object.size = 1
        # There is no error raised here.
        sqlalchemyutils.paginate_query(self.query, self.model, 10,
                                       sort_keys=['display_name',
                                                  'updated_at',
                                                  'size'],
                                       marker=marker_object,
                                       sort_dirs=['desc', 'asc', 'desc'])



#    (c) Copyright 2014 Brocade Communications Systems Inc.
#    All Rights Reserved.
#
#    Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""
ZoneManager is responsible to manage access control using FC zoning
when zoning mode is set as 'fabric'.
ZoneManager provides interfaces to add connection and remove connection
for given initiator and target list associated with a FC volume attach and
detach operation.

**Related Flags**

:zone_driver:  Used by:class:`ZoneManager`.
    Defaults to
    `cinder.zonemanager.drivers.brocade.brcd_fc_zone_driver.BrcdFCZoneDriver`
:zoning_policy: Used by: class: 'ZoneManager'. Defaults to 'none'

"""

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils
import six

from cinder import exception
from cinder.i18n import _
from cinder.volume import configuration as config
from cinder.zonemanager import fc_common
import cinder.zonemanager.fczm_constants as zone_constant


LOG = logging.getLogger(__name__)

zone_manager_opts = [
    cfg.StrOpt('zone_driver',
               default='cinder.zonemanager.drivers.brocade.brcd_fc_zone_driver'
               '.BrcdFCZoneDriver',
               help='FC Zone Driver responsible for zone management'),
    cfg.StrOpt('zoning_policy',
               default='initiator-target',
               help='Zoning policy configured by user; valid values include '
               '"initiator-target" or "initiator"'),
    cfg.StrOpt('fc_fabric_names',
               help='Comma separated list of Fibre Channel fabric names.'
               ' This list of names is used to retrieve other SAN credentials'
               ' for connecting to each SAN fabric'),
    cfg.StrOpt('fc_san_lookup_service',
               default='cinder.zonemanager.drivers.brocade'
               '.brcd_fc_san_lookup_service.BrcdFCSanLookupService',
               help='FC SAN Lookup Service'),
    cfg.BoolOpt('enable_unsupported_driver',
                default=False,
                help="Set this to True when you want to allow an unsupported "
                     "zone manager driver to start.  Drivers that haven't "
                     "maintained a working CI system and testing are marked "
                     "as unsupported until CI is working again.  This also "
                     "marks a driver as deprecated and may be removed in the "
                     "next release."),

]

CONF = cfg.CONF
CONF.register_opts(zone_manager_opts, group='fc-zone-manager')


class ZoneManager(fc_common.FCCommon):

    """Manages Connection control during attach/detach.

       Version History:
           1.0 - Initial version
           1.0.1 - Added __new__ for singleton
           1.0.2 - Added friendly zone name

    """

    VERSION = "1.0.2"
    driver = None
    _initialized = False
    fabric_names = []

    def __new__(class_, *args, **kwargs):
        if not hasattr(class_, "_instance"):
            class_._instance = object.__new__(class_)
        return class_._instance

    def __init__(self, **kwargs):
        """Load the driver from the one specified in args, or from flags."""
        super(ZoneManager, self).__init__(**kwargs)

        self.configuration = config.Configuration(zone_manager_opts,
                                                  'fc-zone-manager')
        self.set_initialized(False)
        self._build_driver()

    def _build_driver(self):
        zone_driver = self.configuration.zone_driver
        LOG.debug("Zone driver from config: %(driver)s",
                  {'driver': zone_driver})

        zm_config = config.Configuration(zone_manager_opts, 'fc-zone-manager')
        # Initialize vendor specific implementation of FCZoneDriver
        self.driver = importutils.import_object(
            zone_driver,
            configuration=zm_config)

        if not self.driver.supported:
            self._log_unsupported_driver_warning()

            if not self.configuration.enable_unsupported_driver:
                LOG.error("Unsupported drivers are disabled."
                          " You can re-enable by adding "
                          "enable_unsupported_driver=True to the "
                          "fc-zone-manager section in cinder.conf",
                          resource={'type': 'zone_manager',
                                    'id': self.__class__.__name__})
                return

        self.set_initialized(True)

    @property
    def initialized(self):
        return self._initialized

    def set_initialized(self, value=True):
        self._initialized = value

    def _require_initialized(self):
        """Verifies that the zone manager has been properly initialized."""
        if not self.initialized:
            LOG.error("Fibre Channel Zone Manager is not initialized.")
            raise exception.ZoneManagerNotInitialized()
        else:
            self._log_unsupported_driver_warning()

    def _log_unsupported_driver_warning(self):
        """Annoy the log about unsupported fczm drivers."""
        if not self.driver.supported:
            LOG.warning("Zone Manager driver (%(driver_name)s %(version)s)"
                        " is currently unsupported and may be removed in "
                        "the next release of OpenStack. Use at your own "
                        "risk.",
                        {'driver_name': self.driver.__class__.__name__,
                         'version': self.driver.get_version()},
                        resource={'type': 'zone_manager',
                                  'id': self.driver.__class__.__name__})

    def get_zoning_state_ref_count(self, initiator_wwn, target_wwn):
        """Zone management state check.

        Performs state check for given I-T pair to return the current count of
        active attach for the pair.
        """
        # TODO(sk): ref count state management
        count = 0
        # check the state for I-T pair
        return count

    def add_connection(self, conn_info):
        """Add connection control.

        Adds connection control for the given initiator target map.
        initiator_target_map - each initiator WWN mapped to a list of one
        or more target WWN:

        .. code-block:: python

            e.g.:
            {
                '10008c7cff523b01': ['20240002ac000a50', '20240002ac000a40']
            }
        """
        connected_fabric = None
        host_name = None
        storage_system = None

        try:
            # Make sure the driver is loaded and we are initialized
            self._log_unsupported_driver_warning()
            self._require_initialized()
        except exception.ZoneManagerNotInitialized:
            LOG.error("Cannot add Fibre Channel Zone because the "
                      "Zone Manager is not initialized properly.",
                      resource={'type': 'zone_manager',
                                'id': self.__class__.__name__})
            return

        try:
            initiator_target_map = (
                conn_info[zone_constant.DATA][zone_constant.IT_MAP])

            if zone_constant.HOST in conn_info[zone_constant.DATA]:
                host_name = conn_info[
                    zone_constant.DATA][
                    zone_constant.HOST].replace(" ", "_")

            if zone_constant.STORAGE in conn_info[zone_constant.DATA]:
                storage_system = (
                    conn_info[
                        zone_constant.DATA][
                        zone_constant.STORAGE].replace(" ", "_"))

            for initiator in initiator_target_map.keys():
                target_list = initiator_target_map[initiator]
                LOG.debug("Target list : %(targets)s",
                          {'targets': target_list})

                # get SAN context for the target list
                fabric_map = self.get_san_context(target_list)
                LOG.debug("Fabric map after context lookup: %(fabricmap)s",
                          {'fabricmap': fabric_map})
                # iterate over each SAN and apply connection control
                for fabric in fabric_map.keys():
                    connected_fabric = fabric
                    t_list = fabric_map[fabric]
                    # get valid I-T map to add connection control
                    i_t_map = {initiator: t_list}
                    valid_i_t_map = self.get_valid_initiator_target_map(
                        i_t_map, True)
                    LOG.info("Final filtered map for fabric: %(i_t_map)s",
                             {'i_t_map': valid_i_t_map})

                    # Call driver to add connection control
                    self.driver.add_connection(fabric, valid_i_t_map,
                                               host_name, storage_system)

            LOG.info("Add connection: finished iterating "
                     "over all target list")
        except Exception as e:
            msg = _("Failed adding connection for fabric=%(fabric)s: "
                    "Error: %(err)s") % {'fabric': connected_fabric,
                                         'err': six.text_type(e)}
            LOG.error(msg)
            raise exception.ZoneManagerException(reason=msg)

    def delete_connection(self, conn_info):
        """Delete connection.

        Updates/deletes connection control for the given initiator target map.
        initiator_target_map - each initiator WWN mapped to a list of one
        or more target WWN:

        .. code-block:: python

            e.g.:
            {
                '10008c7cff523b01': ['20240002ac000a50', '20240002ac000a40']
            }
        """
        connected_fabric = None
        host_name = None
        storage_system = None

        try:
            # Make sure the driver is loaded and we are initialized
            self._log_unsupported_driver_warning()
            self._require_initialized()
        except exception.ZoneManagerNotInitialized:
            LOG.error("Cannot delete fibre channel zone because the "
                      "Zone Manager is not initialized properly.",
                      resource={'type': 'zone_manager',
                                'id': self.__class__.__name__})
            return

        try:
            initiator_target_map = (
                conn_info[zone_constant.DATA][zone_constant.IT_MAP])

            if zone_constant.HOST in conn_info[zone_constant.DATA]:
                host_name = conn_info[zone_constant.DATA][zone_constant.HOST]

            if zone_constant.STORAGE in conn_info[zone_constant.DATA]:
                storage_system = (
                    conn_info[
                        zone_constant.DATA][
                        zone_constant.STORAGE].replace(" ", "_"))

            for initiator in initiator_target_map.keys():
                target_list = initiator_target_map[initiator]
                LOG.info("Delete connection target list: %(targets)s",
                         {'targets': target_list})

                # get SAN context for the target list
                fabric_map = self.get_san_context(target_list)
                LOG.debug("Delete connection fabric map from SAN "
                          "context: %(fabricmap)s", {'fabricmap': fabric_map})

                # iterate over each SAN and apply connection control
                for fabric in fabric_map.keys():
                    connected_fabric = fabric
                    t_list = fabric_map[fabric]
                    # get valid I-T map to add connection control
                    i_t_map = {initiator: t_list}
                    valid_i_t_map = self.get_valid_initiator_target_map(
                        i_t_map, False)
                    LOG.info("Final filtered map for delete connection: "
                             "%(i_t_map)s", {'i_t_map': valid_i_t_map})

                    # Call driver to delete connection control
                    if len(valid_i_t_map) > 0:
                        self.driver.delete_connection(fabric,
                                                      valid_i_t_map,
                                                      host_name,
                                                      storage_system)

            LOG.debug("Delete connection - finished iterating over all"
                      " target list")
        except Exception as e:
            msg = _("Failed removing connection for fabric=%(fabric)s: "
                    "Error: %(err)s") % {'fabric': connected_fabric,
                                         'err': six.text_type(e)}
            LOG.error(msg)
            raise exception.ZoneManagerException(reason=msg)

    def get_san_context(self, target_wwn_list):
        """SAN lookup for end devices.

        Look up each SAN configured and return a map of SAN (fabric IP)
        to list of target WWNs visible to the fabric.
        """
        fabric_map = self.driver.get_san_context(target_wwn_list)
        LOG.debug("Got SAN context: %(fabricmap)s", {'fabricmap': fabric_map})
        return fabric_map

    def get_valid_initiator_target_map(self, initiator_target_map,
                                       add_control):
        """Reference count check for end devices.

        Looks up the reference count for each initiator-target pair from the
        map and returns a filtered list based on the operation type
        add_control - operation type can be true for add connection control
        and false for remove connection control
        """
        filtered_i_t_map = {}
        for initiator in initiator_target_map.keys():
            t_list = initiator_target_map[initiator]
            for target in t_list:
                count = self.get_zoning_state_ref_count(initiator, target)
                if add_control:
                    if count > 0:
                        t_list.remove(target)
                    # update count = count + 1
                else:
                    if count > 1:
                        t_list.remove(target)
                    # update count = count - 1
            if t_list:
                filtered_i_t_map[initiator] = t_list
            else:
                LOG.info("No targets to add or remove connection for "
                         "initiator: %(init_wwn)s",
                         {'init_wwn': initiator})
        return filtered_i_t_map



# Copyright (c) 2019 FUJITSU LIMITED
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Cinder Volume driver for Fujitsu ETERNUS DX S3 series."""
import six

from cinder.i18n import _
from cinder import ssh_utils


class FJDXCLI(object):
    """ETERNUS CLI Code."""

    def __init__(self, user, storage_ip, password=None, keyfile=None):
        """Constructor."""
        self.user = user
        self.storage_ip = storage_ip
        if password and keyfile:
            raise Exception(_('can not specify both password and keyfile'))

        self.use_ipv6 = False
        if storage_ip.find(':') != -1:
            self.use_ipv6 = True

        if password:
            self.ssh_pool = ssh_utils.SSHPool(storage_ip, 22, None, user,
                                              password=password, max_size=2)

        if keyfile:
            self.ssh_pool = ssh_utils.SSHPool(storage_ip, 22, None, user,
                                              privatekey=keyfile, max_size=2)

        self.ce_support = False
        self.CMD_dic = {
            'check_user_role': self._check_user_role,
            'show_pool_provision': self._show_pool_provision,
        }

        self.SMIS_dic = {
            '0000': '0',  # Success.
            '0060': '32787',  # The device is in busy state.
            '0100': '4097'
        }  # Size not supported.

    def done(self, command, **option):
        func = self.CMD_dic.get(command, self._default_func)
        return func(**option)

    def _exec_cli(self, cmd, StrictHostKeyChecking=True, **option):

        exec_cmdline = cmd + self._get_option(**option)
        stdoutdata = self._exec_cli_with_eternus(exec_cmdline)
        output = []
        message = []
        stdoutlist = stdoutdata.split('\r\n')
        output_header = ""

        for no, outline in enumerate(stdoutlist):
            if len(outline) <= 0 or outline is None:
                continue

            if not output_header.endswith(exec_cmdline):
                output_header += outline
                continue

            if 0 <= outline.find('Error'):
                raise Exception(_("Output: %(outline)s: "
                                  "Command: %(cmdline)s")
                                % {'outline': outline,
                                   'cmdline': exec_cmdline})

            if not self._is_status(outline):
                continue

            status = int(outline, 16)
            lineno = no + 1
            break
        else:
            raise Exception(_(
                "Invalid CLI output: %(exec_cmdline)s, %(stdoutlist)s")
                % {'exec_cmdline': exec_cmdline,
                   'stdoutlist': stdoutlist})

        if status == 0:
            rc = '0'
            for outline in stdoutlist[lineno:]:
                if 0 <= outline.find('CLI>'):
                    continue
                if len(outline) <= 0:
                    continue
                if outline is None:
                    continue
                message.append(outline)
        else:
            code = stdoutlist[lineno]
            for outline in stdoutlist[lineno + 1:]:
                if 0 <= outline.find('CLI>'):
                    continue
                if len(outline) <= 0:
                    continue
                if outline is None:
                    continue
                output.append(outline)

            rc, message = self._create_error_message(code, output)

        return {'result': 0, 'rc': rc, 'message': message}

    def _exec_cli_with_eternus(self, exec_cmdline):
        """Execute CLI command with arguments."""
        ssh = None
        try:
            ssh = self.ssh_pool.get()
            chan = ssh.invoke_shell()
            chan.send(exec_cmdline + '\n')
            stdoutdata = ''
            while True:
                temp = chan.recv(65535)
                if isinstance(temp, six.binary_type):
                    temp = temp.decode('utf-8')
                else:
                    temp = str(temp)
                stdoutdata += temp

                # CLI command end with 'CLI>'.
                if stdoutdata == '\r\nCLI> ':
                    continue
                if (stdoutdata[len(stdoutdata) - 5: len(stdoutdata) - 1] ==
                        'CLI>'):
                    break
        except Exception as e:
            raise Exception(_("Execute CLI "
                              "command error. Error: %s") % six.text_type(e))
        finally:
            if ssh:
                self.ssh_pool.put(ssh)
                self.ssh_pool.remove(ssh)
        return stdoutdata

    def _create_error_message(self, code, msg):
        """Create error code and message using arguements."""
        message = None
        if code in self.SMIS_dic:
            rc = self.SMIS_dic[code]
        else:
            rc = 'E' + code

            # TODO(whfnst): we will have a dic to store errors.
            if rc == "E0001":
                message = "Bad value: %s" % msg
            elif rc == "ED184":
                message = "Because OPC is being executed, "
                "the processing was discontinued."
            else:
                message = msg

        return rc, message

    @staticmethod
    def _is_status(value):
        """Check whether input value is status value or not."""
        try:
            if len(value) != 2:
                return False

            int(value, 16)
            int(value[0], 16)
            int(value[1], 16)

            return True
        except ValueError:
            return False

    @staticmethod
    def _get_option(**option):
        """Create option strings from dictionary."""
        ret = ""
        for key, value in six.iteritems(option):
            ret += " -%(key)s %(value)s" % {'key': key, 'value': value}
        return ret

    def _default_func(self, **option):
        """Default function."""
        raise Exception(_("Invalid function is specified"))

    def _check_user_role(self, **option):
        """Check user role."""
        try:
            output = self._exec_cli("show users",
                                    StrictHostKeyChecking=False,
                                    **option)
            # Return error.
            rc = output['rc']
            if rc != "0":
                return output

            userlist = output.get('message')
            role = None
            for userinfo in userlist:
                username = userinfo.split('\t')[0]
                if username == self.user:
                    role = userinfo.split('\t')[1]
                    break

            output['message'] = role
        except Exception as ex:
            if 'show users' in six.text_type(ex):
                msg = ("Specified user(%s) does not have Software role"
                       % self.user)
            elif 'Error connecting' in six.text_type(ex):
                msg = (six.text_type(ex)[34:] +
                       ', Please check fujitsu_private_key_path or .xml file')
            else:
                msg = six.text_type(ex)
            output = {
                'result': 0,
                'rc': '4',
                'message': msg
            }
        return output

    def _show_pool_provision(self, **option):
        """Get TPP provision capacity information."""
        try:
            output = self._exec_cli("show volumes", **option)

            rc = output['rc']

            if rc != "0":
                return output

            clidatalist = output.get('message')

            data = 0
            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                if clidata[0] == 'FFFF':
                    break
                data += int(clidata[7], 16)
            provision = data / 2097152

            output['message'] = provision
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show pool provision capacity error: %s"
                           % six.text_type(ex)
            }

        return output



# Copyright 2020 Datera
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import ipaddress
import math
import random
import time
import uuid

import eventlet
from os_brick import exception as brick_exception
from oslo_log import log as logging
from oslo_serialization import jsonutils as json
from oslo_utils import importutils
from oslo_utils import units
import six

from cinder import exception
from cinder.i18n import _
from cinder.image import image_utils
from cinder import utils
import cinder.volume.drivers.datera.datera_common as datc
from cinder.volume import volume_types
from cinder.volume import volume_utils as volutils

LOG = logging.getLogger(__name__)

dexceptions = importutils.try_import('dfs_sdk.exceptions')

API_VERSION = "2.1"


# The DateraAPI classes (2.1, 2.2) are enhanced by datera_common's lookup()
# decorator which generates members run-time. Therefore on the class we disable
# pylint's no-member check pylint: disable=no-member
class DateraApi(object):

    # =================
    # = Create Volume =
    # =================

    def _create_volume_2_1(self, volume):
        policies = self._get_policies_for_resource(volume)
        num_replicas = int(policies['replica_count'])
        storage_name = 'storage-1'
        volume_name = 'volume-1'
        template = policies['template']
        placement = policies['placement_mode']
        ip_pool = policies['ip_pool']

        name = datc.get_name(volume)

        if template:
            app_params = (
                {
                    'create_mode': 'openstack',
                    # 'uuid': str(volume['id']),
                    'name': name,
                    'app_template': {'path': '/app_templates/{}'.format(
                        template)}
                })
        else:

            app_params = (
                {
                    'create_mode': 'openstack',
                    'uuid': str(volume['id']),
                    'name': name,
                    'access_control_mode': 'deny_all',
                    'storage_instances': [
                        {
                            'name': storage_name,
                            'ip_pool': {'path': ('/access_network_ip_pools/'
                                                 '{}'.format(ip_pool))},
                            'volumes': [
                                {
                                    'name': volume_name,
                                    'size': volume['size'],
                                    'placement_mode': placement,
                                    'replica_count': num_replicas,
                                    'snapshot_policies': [
                                    ]
                                }
                            ]
                        }
                    ]
                })

        tenant = self.create_tenant(volume['project_id'])
        self.api.app_instances.create(tenant=tenant, **app_params)
        self._update_qos_2_1(volume, policies)
        self._add_vol_meta_2_1(volume)

    # =================
    # = Extend Volume =
    # =================

    def _extend_volume_2_1(self, volume, new_size):
        if volume['size'] >= new_size:
            LOG.warning("Volume size not extended due to original size being "
                        "greater or equal to new size.  Originial: "
                        "%(original)s, New: %(new)s", {
                            'original': volume['size'],
                            'new': new_size})
            return
        policies = self._get_policies_for_resource(volume)
        template = policies['template']
        if template:
            LOG.warning("Volume size not extended due to template binding."
                        " volume: %(volume)s, template: %(template)s",
                        {'volume': volume, 'template': template})
            return

        with self._offline_flip_2_1(volume):
            # Change Volume Size
            tenant = self.get_tenant(volume['project_id'])
            dvol = self.cvol_to_dvol(volume, tenant=tenant)
            dvol.set(tenant=tenant, size=new_size)

    # =================
    # = Cloned Volume =
    # =================

    def _create_cloned_volume_2_1(self, volume, src_vref):
        tenant = self.get_tenant(volume['project_id'])
        sdvol = self.cvol_to_dvol(src_vref, tenant=tenant)
        src = sdvol.path
        data = {
            'create_mode': 'openstack',
            'name': datc.get_name(volume),
            'uuid': str(volume['id']),
            'clone_volume_src': {'path': src},
        }
        tenant = self.get_tenant(volume['project_id'])
        self.api.app_instances.create(tenant=tenant, **data)

        if volume['size'] > src_vref['size']:
            self._extend_volume_2_1(volume, volume['size'])
        self._add_vol_meta_2_1(volume)

    # =================
    # = Delete Volume =
    # =================

    def _delete_volume_2_1(self, volume):
        try:
            tenant = self.get_tenant(volume['project_id'])
            ai = self.cvol_to_ai(volume, tenant=tenant)
            si = ai.storage_instances.list(tenant=tenant)[0]

            # Clear out ACL
            acl = si.acl_policy.get(tenant=tenant)
            acl.set(tenant=tenant, initiators=[])

            # Bring volume offline
            data = {
                'admin_state': 'offline',
                'force': True
            }
            ai.set(tenant=tenant, **data)

            ai.delete(tenant=tenant, force=True)
        except exception.NotFound:
            msg = ("Tried to delete volume %s, but it was not found in the "
                   "Datera cluster. Continuing with delete.")
            LOG.info(msg, datc.get_name(volume))

    # =================
    # = Ensure Export =
    # =================

    def _ensure_export_2_1(self, context, volume, connector=None):
        pass

    # =========================
    # = Initialize Connection =
    # =========================

    def _initialize_connection_2_1(self, volume, connector):
        # Now online the app_instance (which will online all storage_instances)
        multipath = connector.get('multipath', False)
        tenant = self.get_tenant(volume['project_id'])
        ai = self.cvol_to_ai(volume, tenant=tenant)
        data = {
            'admin_state': 'online'
        }
        ai.set(tenant=tenant, **data)
        si = ai.storage_instances.list(tenant=tenant)[0]

        # randomize portal chosen
        choice = 0
        policies = self._get_policies_for_resource(volume)
        if policies["round_robin"]:
            choice = random.randint(0, 1)
        portal = si.access['ips'][choice] + ':3260'
        iqn = si.access['iqn']
        if multipath:
            portals = [p + ':3260' for p in si.access['ips']]
            iqns = [iqn for _ in si.access['ips']]
            lunids = [self._get_lunid() for _ in si.access['ips']]

            result = {
                'driver_volume_type': 'iscsi',
                'data': {
                    'target_discovered': False,
                    'target_iqn': iqn,
                    'target_iqns': iqns,
                    'target_portal': portal,
                    'target_portals': portals,
                    'target_lun': self._get_lunid(),
                    'target_luns': lunids,
                    'volume_id': volume['id'],
                    'discard': False}}
        else:
            result = {
                'driver_volume_type': 'iscsi',
                'data': {
                    'target_discovered': False,
                    'target_iqn': iqn,
                    'target_portal': portal,
                    'target_lun': self._get_lunid(),
                    'volume_id': volume['id'],
                    'discard': False}}

        if self.use_chap_auth:
            result['data'].update(
                auth_method="CHAP",
                auth_username=self.chap_username,
                auth_password=self.chap_password)

        return result

    # =================
    # = Create Export =
    # =================

    def _create_export_2_1(self, context, volume, connector):
        tenant = self.get_tenant(volume['project_id'])
        ai = self.cvol_to_ai(volume, tenant=tenant)
        data = {
            'admin_state': 'offline',
            'force': True
        }
        ai.set(tenant=tenant, **data)
        si = ai.storage_instances.list(tenant=tenant)[0]
        policies = self._get_policies_for_resource(volume)
        if connector and connector.get('ip'):
            # Case where volume_type has non default IP Pool info
            if policies['ip_pool'] != 'default':
                initiator_ip_pool_path = self.api.access_network_ip_pools.get(
                    policies['ip_pool']).path
            # Fallback to trying reasonable IP based guess
            else:
                initiator_ip_pool_path = self._get_ip_pool_for_string_ip_2_1(
                    connector['ip'], tenant)

            ip_pool_data = {'ip_pool': {'path': initiator_ip_pool_path}}
            si.set(tenant=tenant, **ip_pool_data)
        data = {
            'admin_state': 'online'
        }
        ai.set(tenant=tenant, **data)
        # Check if we've already setup everything for this volume
        storage_instances = ai.storage_instances.list(tenant=tenant)
        # Handle adding initiator to product if necessary
        # Then add initiator to ACL
        if connector and connector.get('initiator'):
            initiator_name = "OpenStack-{}".format(str(uuid.uuid4())[:8])
            initiator = connector['initiator']
            dinit = None
            data = {'id': initiator, 'name': initiator_name}
            # Try and create the initiator
            # If we get a conflict, ignore it
            try:
                dinit = self.api.initiators.create(tenant=tenant, **data)
            except dexceptions.ApiConflictError:
                dinit = self.api.initiators.get(initiator, tenant=tenant)
            initiator_path = dinit['path']
            # Create ACL with initiator group as reference for each
            # storage_instance in app_instance
            # TODO(_alastor_): We need to avoid changing the ACLs if the
            # template already specifies an ACL policy.
            for si in storage_instances:
                existing_acl = si.acl_policy.get(tenant=tenant)
                data = {}
                # Grabbing only the 'path' key from each existing initiator
                # within the existing acl. eacli --> existing acl initiator
                eacli = []
                for acl in existing_acl['initiators']:
                    nacl = {}
                    nacl['path'] = acl['path']
                    eacli.append(nacl)
                data['initiators'] = eacli
                data['initiators'].append({"path": initiator_path})
                # Grabbing only the 'path' key from each existing initiator
                # group within the existing acl. eaclig --> existing
                # acl initiator group
                eaclig = []
                for acl in existing_acl['initiator_groups']:
                    nacl = {}
                    nacl['path'] = acl['path']
                    eaclig.append(nacl)
                data['initiator_groups'] = eaclig
                si.acl_policy.set(tenant=tenant, **data)
        if self.use_chap_auth:
            for si in storage_instances:
                data = {'type': 'chap',
                        'target_user_name': self.chap_username,
                        'target_pswd': self.chap_password}
                si.auth.set(tenant=tenant, **data)
        # Check to ensure we're ready for go-time
        self._si_poll_2_1(volume, si, tenant)
        self._add_vol_meta_2_1(volume, connector=connector)

    # =================
    # = Detach Volume =
    # =================

    def _detach_volume_2_1(self, context, volume, attachment=None):
        try:
            tenant = self.get_tenant(volume['project_id'])
            ai = self.cvol_to_ai(volume, tenant=tenant)
            # Clear out ACL for this specific attachment
            si = ai.storage_instances.list(tenant=tenant)[0]
            existing_acl = si.acl_policy.get(tenant=tenant)
            data = {}
            # Grabbing only the 'path' key from each existing initiator
            # within the existing acl. eacli --> existing acl initiator
            eacli = []
            for acl in existing_acl['initiators']:
                if (
                    attachment is not None
                    and attachment.connector is not None
                    and acl['path'].split('/')[-1]
                    == attachment.connector['initiator']
                ):
                    continue
                nacl = {}
                nacl['path'] = acl['path']
                eacli.append(nacl)
            data['initiators'] = eacli
            data['initiator_groups'] = existing_acl['initiator_groups']
            si.acl_policy.set(tenant=tenant, **data)

            if not eacli:
                # bring the application instance offline if there
                # are no initiators left.
                data = {
                    'admin_state': 'offline',
                    'force': True
                }
                ai.set(tenant=tenant, **data)

        except exception.NotFound:
            msg = ("Tried to detach volume %s, but it was not found in the "
                   "Datera cluster. Continuing with detach.")
            LOG.info(msg, volume['id'])

    # ===================
    # = Create Snapshot =
    # ===================

    def _create_snapshot_2_1(self, snapshot):

        dummy_vol = {'id': snapshot['volume_id'],
                     'project_id': snapshot['project_id']}
        tenant = self.get_tenant(dummy_vol['project_id'])
        dvol = self.cvol_to_dvol(dummy_vol, tenant=tenant)
        snap_params = {
            'uuid': snapshot['id'],
        }
        snap = dvol.snapshots.create(tenant=tenant, **snap_params)
        self._snap_poll_2_1(snap, tenant)

    # ===================
    # = Delete Snapshot =
    # ===================

    def _delete_snapshot_2_1(self, snapshot):
        # Handle case where snapshot is "managed"
        dummy_vol = {'id': snapshot['volume_id'],
                     'project_id': snapshot['project_id']}
        tenant = self.get_tenant(dummy_vol['project_id'])
        dvol = self.cvol_to_dvol(dummy_vol, tenant=tenant)

        snapshots = None

        # Shortcut if this is a managed snapshot
        provider_location = snapshot.get('provider_location')
        if provider_location:
            snap = dvol.snapshots.get(provider_location, tenant=tenant)
            snap.delete(tenant=tenant)
            return

        # Long-way.  UUID identification
        try:
            snapshots = dvol.snapshots.list(tenant=tenant)
        except exception.NotFound:
            msg = ("Tried to delete snapshot %s, but parent volume %s was "
                   "not found in Datera cluster. Continuing with delete.")
            LOG.info(msg,
                     datc.get_name(snapshot),
                     datc.get_name({'id': snapshot['volume_id']}))
            return

        try:
            for snap in snapshots:
                if snap.uuid == snapshot['id']:
                    snap.delete(tenant=tenant)
                    break
            else:
                raise exception.NotFound
        except exception.NotFound:
            msg = ("Tried to delete snapshot %s, but was not found in "
                   "Datera cluster. Continuing with delete.")
            LOG.info(msg, datc.get_name(snapshot))

    # ========================
    # = Volume From Snapshot =
    # ========================

    def _create_volume_from_snapshot_2_1(self, volume, snapshot):
        # Handle case where snapshot is "managed"
        dummy_vol = {'id': snapshot['volume_id'],
                     'project_id': snapshot['project_id']}
        tenant = self.get_tenant(dummy_vol['project_id'])
        dvol = self.cvol_to_dvol(dummy_vol, tenant=tenant)
        found_snap = None
        provider_location = snapshot.get('provider_location')
        if provider_location:
            found_snap = dvol.snapshots.get(provider_location, tenant=tenant)
        else:
            snapshots = dvol.snapshots.list(tenant=tenant)
            for snap in snapshots:
                if snap.uuid == snapshot['id']:
                    found_snap = snap
                    break
            else:
                raise exception.SnapshotNotFound(snapshot_id=snapshot['id'])

        self._snap_poll_2_1(found_snap, tenant)

        src = found_snap.path
        app_params = (
            {
                'create_mode': 'openstack',
                'uuid': str(volume['id']),
                'name': datc.get_name(volume),
                'clone_snapshot_src': {'path': src},
            })

        self.api.app_instances.create(tenant=tenant, **app_params)
        if (volume['size'] > snapshot['volume_size']):
            self._extend_volume_2_1(volume, volume['size'])
        self._add_vol_meta_2_1(volume)

    # ==========
    # = Retype =
    # ==========

    def _retype_2_1(self, ctxt, volume, new_type, diff, host):
        LOG.debug("Retype called\n"
                  "Volume: %(volume)s\n"
                  "NewType: %(new_type)s\n"
                  "Diff: %(diff)s\n"
                  "Host: %(host)s\n", {'volume': volume, 'new_type': new_type,
                                       'diff': diff, 'host': host})
        # We'll take the fast route only if the types share the same backend
        # And that backend matches this driver
        old_pol = self._get_policies_for_resource(volume)
        new_pol = self._get_policies_for_volume_type(new_type)
        if (host['capabilities']['volume_backend_name'].lower() ==
                self.backend_name.lower()):
            LOG.debug("Starting fast volume retype")

            if old_pol.get('template') or new_pol.get('template'):
                LOG.warning(
                    "Fast retyping between template-backed volume-types "
                    "unsupported.  Type1: %s, Type2: %s",
                    volume['volume_type_id'], new_type)

            self._update_qos_2_1(volume, new_pol, clear_old=True)
            tenant = self.get_tenant(volume['project_id'])
            dvol = self.cvol_to_dvol(volume, tenant=tenant)
            # Only replica_count ip_pool requires offlining the app_instance
            if (new_pol['replica_count'] != old_pol['replica_count'] or
                    new_pol['ip_pool'] != old_pol['ip_pool']):
                with self._offline_flip_2_1(volume):
                    vol_params = (
                        {
                            'placement_mode': new_pol['placement_mode'],
                            'replica_count': new_pol['replica_count'],
                        })
                    dvol.set(tenant=tenant, **vol_params)
            elif new_pol['placement_mode'] != old_pol['placement_mode']:
                vol_params = (
                    {
                        'placement_mode': new_pol['placement_mode'],
                    })
                dvol.set(tenant=tenant, **vol_params)
            self._add_vol_meta_2_1(volume)
            return True

        else:
            LOG.debug("Couldn't fast-retype volume between specified types")
            return False

    # ==========
    # = Manage =
    # ==========

    def _manage_existing_2_1(self, volume, existing_ref):
        # Only volumes created under the requesting tenant can be managed in
        # the v2.1+ API.  Eg.  If tenant A is the tenant for the volume to be
        # managed, it must also be tenant A that makes this request.
        # This will be fixed in a later API update
        existing_ref = existing_ref['source-name']
        app_inst_name, __, __, __ = datc._parse_vol_ref(existing_ref)
        LOG.debug("Managing existing Datera volume %s  "
                  "Changing name to %s",
                  datc.get_name(volume), existing_ref)
        # Rename AppInstance
        dummy_vol = {'id': app_inst_name,
                     'project_id': volume['project_id']}
        tenant = self.get_tenant(volume['project_id'])
        ai = self.cvol_to_ai(dummy_vol, tenant=tenant)
        data = {'name': datc.get_name(volume)}
        ai.set(tenant=tenant, **data)
        self._add_vol_meta_2_1(volume)

    # ===================
    # = Manage Get Size =
    # ===================

    def _manage_existing_get_size_2_1(self, volume, existing_ref):
        existing_ref = existing_ref['source-name']
        app_inst_name, storage_inst_name, vol_name, __ = datc._parse_vol_ref(
            existing_ref)
        dummy_vol = {'id': app_inst_name,
                     'project_id': volume['project_id']}
        dvol = self.cvol_to_dvol(dummy_vol)
        return dvol.size

    # =========================
    # = Get Manageable Volume =
    # =========================

    def _list_manageable_2_1(self, cinder_volumes):
        # Use the first volume to determine the tenant we're working under
        if cinder_volumes:
            tenant = self.get_tenant(cinder_volumes[0]['project_id'])
        else:
            tenant = None
        app_instances = self.api.app_instances.list(tenant=tenant)

        results = []

        if cinder_volumes and 'volume_id' in cinder_volumes[0]:
            cinder_volume_ids = [vol['volume_id'] for vol in cinder_volumes]
        else:
            cinder_volume_ids = [vol['id'] for vol in cinder_volumes]

        for ai in app_instances:
            ai_name = ai['name']
            reference = None
            size = None
            safe_to_manage = False
            reason_not_safe = ""
            cinder_id = None
            extra_info = {}
            (safe_to_manage, reason_not_safe,
                cinder_id) = self._is_manageable_2_1(
                    ai, cinder_volume_ids, tenant)
            si = ai.storage_instances.list(tenant=tenant)[0]
            si_name = si.name
            vol = si.volumes.list(tenant=tenant)[0]
            vol_name = vol.name
            size = vol.size
            snaps = [(snap.utc_ts, snap.uuid)
                     for snap in vol.snapshots.list(tenant=tenant)]
            extra_info["snapshots"] = json.dumps(snaps)
            reference = {"source-name": "{}:{}:{}".format(
                ai_name, si_name, vol_name)}

            results.append({
                'reference': reference,
                'size': size,
                'safe_to_manage': safe_to_manage,
                'reason_not_safe': reason_not_safe,
                'cinder_id': cinder_id,
                'extra_info': extra_info})
        return results

    def _get_manageable_volumes_2_1(self, cinder_volumes, marker, limit,
                                    offset, sort_keys, sort_dirs):
        LOG.debug("Listing manageable Datera volumes")
        results = self._list_manageable_2_1(cinder_volumes)
        page_results = volutils.paginate_entries_list(
            results, marker, limit, offset, sort_keys, sort_dirs)

        return page_results

    def _is_manageable_2_1(self, ai, cinder_volume_ids, tenant):
        cinder_id = None
        ai_name = ai.name
        match = datc.UUID4_RE.match(ai_name)
        if match:
            cinder_id = match.group(1)
        if cinder_id and cinder_id in cinder_volume_ids:
            return (False,
                    "App Instance already managed by Cinder",
                    cinder_id)
        if len(ai.storage_instances.list(tenant=tenant)) == 1:
            si = ai.storage_instances.list(tenant=tenant)[0]
            if len(si['volumes']) == 1:
                return (True, "", cinder_id)
        return (False,
                "App Instance has more than one storage instance or volume",
                cinder_id)
    # ============
    # = Unmanage =
    # ============

    def _unmanage_2_1(self, volume):
        LOG.debug("Unmanaging Cinder volume %s.  Changing name to %s",
                  volume['id'], datc.get_unmanaged(volume['id']))
        data = {'name': datc.get_unmanaged(volume['id'])}
        tenant = self.get_tenant(volume['project_id'])
        ai = self.cvol_to_ai(volume, tenant=tenant)
        ai.set(tenant=tenant, **data)

    # ===================
    # = Manage Snapshot =
    # ===================

    def _manage_existing_snapshot_2_1(self, snapshot, existing_ref):
        existing_ref = existing_ref['source-name']
        datc._check_snap_ref(existing_ref)
        LOG.debug("Managing existing Datera volume snapshot %s for volume %s",
                  existing_ref, datc.get_name({'id': snapshot['volume_id']}))
        return {'provider_location': existing_ref}

    def _manage_existing_snapshot_get_size_2_1(self, snapshot, existing_ref):
        existing_ref = existing_ref['source-name']
        datc._check_snap_ref(existing_ref)
        dummy_vol = {'id': snapshot['volume_id'],
                     'project_id': snapshot['project_id']}
        dvol = self.cvol_to_dvol(dummy_vol)
        return dvol.size

    def _get_manageable_snapshots_2_1(self, cinder_snapshots, marker, limit,
                                      offset, sort_keys, sort_dirs):
        LOG.debug("Listing manageable Datera snapshots")
        results = self._list_manageable_2_1(cinder_snapshots)
        snap_results = []
        snapids = set((snap['id'] for snap in cinder_snapshots))
        snaprefs = set((snap.get('provider_location')
                        for snap in cinder_snapshots))
        for volume in results:
            snaps = json.loads(volume["extra_info"]["snapshots"])
            for snapshot in snaps:
                reference = snapshot[0]
                uuid = snapshot[1]
                size = volume["size"]
                safe_to_manage = True
                reason_not_safe = ""
                cinder_id = ""
                extra_info = {}
                source_reference = volume["reference"]
                if uuid in snapids or reference in snaprefs:
                    safe_to_manage = False
                    reason_not_safe = _("already managed by Cinder")
                elif not volume['safe_to_manage'] and not volume['cinder_id']:
                    safe_to_manage = False
                    reason_not_safe = _("parent volume not safe to manage")
                snap_results.append({
                    'reference': {'source-name': reference},
                    'size': size,
                    'safe_to_manage': safe_to_manage,
                    'reason_not_safe': reason_not_safe,
                    'cinder_id': cinder_id,
                    'extra_info': extra_info,
                    'source_reference': source_reference})
        page_results = volutils.paginate_entries_list(
            snap_results, marker, limit, offset, sort_keys, sort_dirs)

        return page_results

    def _unmanage_snapshot_2_1(self, snapshot):
        return {'provider_location': None}

    # ====================
    # = Fast Image Clone =
    # ====================

    def _clone_image_2_1(self, context, volume, image_location, image_meta,
                         image_service):
        # We're not going to fast image clone if the feature is not enabled
        # and/or we can't reach the image being requested
        if (not self.image_cache or
                not self._image_accessible(context, volume, image_meta)):
            return None, False
        # Check to make sure we're working with a valid volume type
        try:
            found = volume_types.get_volume_type(context, self.image_type)
        except (exception.VolumeTypeNotFound, exception.InvalidVolumeType):
            found = None
        if not found:
            msg = "Invalid volume type: %s"
            LOG.error(msg, self.image_type)
            raise ValueError(_("Option datera_image_cache_volume_type_id must"
                               " be set to a valid volume_type id"))
        # Check image format
        fmt = image_meta.get('disk_format', '')
        if fmt.lower() != 'raw':
            LOG.debug("Image format is not RAW, image requires conversion "
                      "before clone.  Image format: [%s]", fmt)
            return None, False

        LOG.debug("Starting fast image clone")
        # TODO(_alastor_): determine if Datera is already an image backend
        # for this request and direct clone instead of caching

        # Dummy volume, untracked by Cinder
        src_vol = {'id': image_meta['id'],
                   'volume_type_id': self.image_type,
                   'size': volume['size'],
                   'project_id': volume['project_id']}

        # Determine if we have a cached version of the image
        cached = self._vol_exists_2_1(src_vol)

        if cached:
            tenant = self.get_tenant(src_vol['project_id'])
            ai = self.cvol_to_ai(src_vol, tenant=tenant)
            metadata = ai.metadata.get(tenant=tenant)
            # Check to see if the master image has changed since we created
            # The cached version
            ts = self._get_vol_timestamp_2_1(src_vol)
            mts = time.mktime(image_meta['updated_at'].timetuple())
            LOG.debug("Original image timestamp: %s, cache timestamp %s",
                      mts, ts)
            # If the image is created by Glance, we'll trust that even if the
            # timestamps don't match up, the data is ok to clone as it's not
            # managed by this driver
            if metadata.get('type') == 'image':
                LOG.debug("Found Glance volume-backed image for %s",
                          src_vol['id'])
            # If the master image time is greater than the volume creation
            # time, we invalidate the cache and delete the volume.  The
            # exception is if the cached volume was created by Glance.  We
            # NEVER want to delete this volume.  It's annotated with
            # 'type': 'image' in the metadata, so we'll check for that
            elif mts > ts and metadata.get('type') != 'image':
                LOG.debug("Cache is older than original image, deleting cache")
                cached = False
                self._delete_volume_2_1(src_vol)

        # If we don't have the image, we'll cache it
        if not cached:
            LOG.debug("No image cache found for: %s, caching image",
                      image_meta['id'])
            self._cache_vol_2_1(context, src_vol, image_meta, image_service)

        # Now perform the clone of the found image or newly cached image
        self._create_cloned_volume_2_1(volume, src_vol)
        # Force volume resize
        vol_size = volume['size']
        volume['size'] = 0
        self._extend_volume_2_1(volume, vol_size)
        volume['size'] = vol_size
        # Determine if we need to retype the newly created volume
        vtype_id = volume.get('volume_type_id')
        if vtype_id and self.image_type and vtype_id != self.image_type:
            vtype = volume_types.get_volume_type(context, vtype_id)
            LOG.debug("Retyping newly cloned volume from type: %s to type: %s",
                      self.image_type, vtype_id)
            diff, discard = volume_types.volume_types_diff(
                context, self.image_type, vtype_id)
            host = {'capabilities': {'vendor_name': self.backend_name}}
            self._retype_2_1(context, volume, vtype, diff, host)
        return None, True

    def _cache_vol_2_1(self, context, vol, image_meta, image_service):
        image_id = image_meta['id']
        # Pull down image and determine if valid
        with image_utils.TemporaryImages.fetch(image_service,
                                               context,
                                               image_id) as tmp_image:
            data = image_utils.qemu_img_info(tmp_image)
            fmt = data.file_format
            if fmt is None:
                raise exception.ImageUnacceptable(
                    reason=_("'qemu-img info' parsing failed."),
                    image_id=image_id)

            backing_file = data.backing_file
            if backing_file is not None:
                raise exception.ImageUnacceptable(
                    image_id=image_id,
                    reason=_("fmt=%(fmt)s backed by:%(backing_file)s")
                    % {'fmt': fmt, 'backing_file': backing_file, })

            vsize = int(
                math.ceil(float(data.virtual_size) / units.Gi))
            vol['size'] = vsize
            vtype = vol['volume_type_id']
            LOG.info("Creating cached image with volume type: %(vtype)s and "
                     "size %(size)s", {'vtype': vtype, 'size': vsize})
            self._create_volume_2_1(vol)
            with self._connect_vol(context, vol) as device:
                LOG.debug("Moving image %s to volume %s",
                          image_meta['id'], datc.get_name(vol))
                image_utils.convert_image(tmp_image,
                                          device,
                                          'raw',
                                          run_as_root=True)
                LOG.debug("Finished moving image %s to volume %s",
                          image_meta['id'], datc.get_name(vol))
                data = image_utils.qemu_img_info(device, run_as_root=True)
                if data.file_format != 'raw':
                    raise exception.ImageUnacceptable(
                        image_id=image_id,
                        reason=_(
                            "Converted to %(vol_format)s, but format is "
                            "now %(file_format)s") % {
                                'vol_format': 'raw',
                                'file_format': data.file_format})
        # TODO(_alastor_): Remove this snapshot creation when we fix
        # "created_at" attribute in the frontend
        # We don't actually care about the snapshot uuid, we just want
        # a single snapshot
        snapshot = {'id': str(uuid.uuid4()),
                    'volume_id': vol['id'],
                    'project_id': vol['project_id']}
        self._create_snapshot_2_1(snapshot)
        metadata = {'type': 'cached_image'}
        tenant = self.get_tenant(vol['project_id'])
        ai = self.cvol_to_ai(vol, tenant=tenant)
        ai.metadata.set(tenant=tenant, **metadata)
        # Cloning offline AI is ~4 seconds faster than cloning online AI
        self._detach_volume_2_1(None, vol)

    def _get_vol_timestamp_2_1(self, volume):
        tenant = self.get_tenant(volume['project_id'])
        dvol = self.cvol_to_dvol(volume, tenant=tenant)
        snapshots = dvol.snapshots.list(tenant=tenant)
        if len(snapshots) == 1:
            return float(snapshots[0].utc_ts)
        else:
            # We'll return 0 if we find no snapshots (or the incorrect number)
            # to ensure the timestamp comparison with the master copy fails
            # since the master copy will always have a timestamp > 0.
            LOG.debug("Number of snapshots found: %s", len(snapshots))
            return 0

    def _vol_exists_2_1(self, volume):
        LOG.debug("Checking if volume %s exists", volume['id'])
        try:
            ai = self.cvol_to_ai(volume)
            LOG.debug("Volume %s exists", volume['id'])
            return ai
        except exception.NotFound:
            LOG.debug("Volume %s not found", volume['id'])
            return None

    @contextlib.contextmanager
    def _connect_vol(self, context, vol):
        connector = None
        try:
            # Start connection, get the connector object and create the
            # export (ACL, IP-Pools, etc)
            conn = self._initialize_connection_2_1(
                vol, {'multipath': False})
            connector = utils.brick_get_connector(
                conn['driver_volume_type'],
                use_multipath=False,
                device_scan_attempts=10,
                conn=conn)
            connector_info = {'initiator': connector.get_initiator()}
            self._create_export_2_1(None, vol, connector_info)
            retries = 10
            attach_info = conn['data']
            while True:
                try:
                    attach_info.update(
                        connector.connect_volume(conn['data']))
                    break
                except brick_exception.FailedISCSITargetPortalLogin:
                    retries -= 1
                    if not retries:
                        LOG.error("Could not log into portal before end of "
                                  "polling period")
                        raise
                    LOG.debug("Failed to login to portal, retrying")
                    eventlet.sleep(2)
            device_path = attach_info['path']
            yield device_path
        finally:
            # Close target connection
            if connector:
                # Best effort disconnection
                try:
                    connector.disconnect_volume(attach_info, attach_info)
                except Exception:
                    pass

    # ===========
    # = Polling =
    # ===========

    def _snap_poll_2_1(self, snap, tenant):
        eventlet.sleep(datc.DEFAULT_SNAP_SLEEP)
        TIMEOUT = 20
        retry = 0
        poll = True
        while poll and not retry >= TIMEOUT:
            retry += 1
            snap = snap.reload(tenant=tenant)
            if snap.op_state == 'available':
                poll = False
            else:
                eventlet.sleep(1)
        if retry >= TIMEOUT:
            raise exception.VolumeDriverException(
                message=_('Snapshot not ready.'))

    def _si_poll_2_1(self, volume, si, tenant):
        # Initial 4 second sleep required for some Datera versions
        eventlet.sleep(datc.DEFAULT_SI_SLEEP)
        TIMEOUT = 10
        retry = 0
        poll = True
        while poll and not retry >= TIMEOUT:
            retry += 1
            si = si.reload(tenant=tenant)
            if si.op_state == 'available':
                poll = False
            else:
                eventlet.sleep(1)
        if retry >= TIMEOUT:
            raise exception.VolumeDriverException(
                message=_('Resource not ready.'))

    # ================
    # = Volume Stats =
    # ================

    def _get_volume_stats_2_1(self, refresh=False):
        # cluster_stats is defined by datera_iscsi
        # pylint: disable=access-member-before-definition
        if refresh or not self.cluster_stats:
            try:
                LOG.debug("Updating cluster stats info.")

                results = self.api.system.get()

                if 'uuid' not in results:
                    LOG.error(
                        'Failed to get updated stats from Datera Cluster.')

                stats = {
                    'volume_backend_name': self.backend_name,
                    'vendor_name': 'Datera',
                    'driver_version': self.VERSION,
                    'storage_protocol': 'iSCSI',
                    'total_capacity_gb': (
                        int(results.total_capacity) / units.Gi),
                    'free_capacity_gb': (
                        int(results.available_capacity) / units.Gi),
                    'reserved_percentage': 0,
                    'QoS_support': True,
                }

                self.cluster_stats = stats
            except exception.DateraAPIException:
                LOG.error('Failed to get updated stats from Datera cluster.')
        return self.cluster_stats

    # =======
    # = QoS =
    # =======

    def _update_qos_2_1(self, volume, policies, clear_old=False):
        tenant = self.get_tenant(volume['project_id'])
        dvol = self.cvol_to_dvol(volume, tenant=tenant)
        type_id = volume.get('volume_type_id', None)
        if type_id is not None:
            iops_per_gb = int(policies.get('iops_per_gb', 0))
            bandwidth_per_gb = int(policies.get('bandwidth_per_gb', 0))
            # Filter for just QOS policies in result. All of their keys
            # should end with "max"
            fpolicies = {k: int(v) for k, v in
                         policies.items() if k.endswith("max")}
            # Filter all 0 values from being passed
            fpolicies = {k: int(v) for k, v in
                         fpolicies.items() if v > 0}
            # Calculate and set iops/gb and bw/gb, but only if they don't
            # exceed total_iops_max and total_bw_max aren't set since they take
            # priority
            if iops_per_gb:
                ipg = iops_per_gb * volume['size']
                # Not using zero, because zero means unlimited
                im = fpolicies.get('total_iops_max', 1)
                r = ipg
                if ipg > im:
                    r = im
                fpolicies['total_iops_max'] = r
            if bandwidth_per_gb:
                bpg = bandwidth_per_gb * volume['size']
                # Not using zero, because zero means unlimited
                bm = fpolicies.get('total_bandwidth_max', 1)
                r = bpg
                if bpg > bm:
                    r = bm
                fpolicies['total_bandwidth_max'] = r
            if fpolicies or clear_old:
                try:
                    pp = dvol.performance_policy.get(tenant=tenant)
                    pp.delete(tenant=tenant)
                except dexceptions.ApiNotFoundError:
                    LOG.debug("No existing performance policy found")
            if fpolicies:
                dvol.performance_policy.create(tenant=tenant, **fpolicies)

    # ============
    # = IP Pools =
    # ============

    def _get_ip_pool_for_string_ip_2_1(self, ip, tenant):
        """Takes a string ipaddress and return the ip_pool API object dict """
        pool = 'default'
        ip_obj = ipaddress.ip_address(six.text_type(ip))
        ip_pools = self.api.access_network_ip_pools.list(tenant=tenant)
        for ipdata in ip_pools:
            for adata in ipdata['network_paths']:
                if not adata.get('start_ip'):
                    continue
                pool_if = ipaddress.ip_interface(
                    "/".join((adata['start_ip'], str(adata['netmask']))))
                if ip_obj in pool_if.network:
                    pool = ipdata.name
        return self.api.access_network_ip_pools.get(pool, tenant=tenant).path
    # ====================
    # = Volume Migration =
    # ====================

    def _update_migrated_volume_2_1(self, context, volume, new_volume,
                                    volume_status):
        """Rename the newly created volume to the original volume.

        So we can find it correctly.
        """
        tenant = self.get_tenant(new_volume['project_id'])
        ai = self.cvol_to_ai(new_volume, tenant=tenant)
        data = {'name': datc.get_name(volume)}
        ai.set(tenant=tenant, **data)
        return {'_name_id': None}

    @contextlib.contextmanager
    def _offline_flip_2_1(self, volume):
        reonline = False
        tenant = self.get_tenant(volume['project_id'])
        ai = self.cvol_to_ai(volume, tenant=tenant)
        if ai.admin_state == 'online':
            reonline = True
        ai.set(tenant=tenant, admin_state='offline')
        yield
        if reonline:
            ai.set(tenant=tenant, admin_state='online')

    def _add_vol_meta_2_1(self, volume, connector=None):
        if not self.do_metadata:
            return
        metadata = {'host': volume.get('host', ''),
                    'display_name': datc.filter_chars(
                        volume.get('display_name', '')),
                    'bootable': str(volume.get('bootable', False)),
                    'availability_zone': volume.get('availability_zone', '')}
        if connector:
            metadata.update(connector)
        LOG.debug("Adding volume metadata: %s", metadata)
        tenant = self.get_tenant(volume['project_id'])
        ai = self.cvol_to_ai(volume, tenant=tenant)
        ai.metadata.set(tenant=tenant, **metadata)



# Copyright 2015 Yahoo Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from oslo_utils import timeutils
import pytz
import six

from cinder import exception
from cinder import objects
from cinder.objects import fields
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_volume
from cinder.tests.unit import objects as test_objects

fake_consistencygroup = {
    'id': fake.CONSISTENCY_GROUP_ID,
    'user_id': fake.USER_ID,
    'project_id': fake.PROJECT_ID,
    'host': 'fake_host',
    'availability_zone': 'fake_az',
    'name': 'fake_name',
    'description': 'fake_description',
    'volume_type_id': fake.VOLUME_TYPE_ID,
    'status': fields.ConsistencyGroupStatus.CREATING,
    'cgsnapshot_id': fake.CGSNAPSHOT_ID,
    'source_cgid': None,
}

fake_cgsnapshot = {
    'id': fake.CGSNAPSHOT_ID,
    'user_id': fake.USER_ID,
    'project_id': fake.PROJECT_ID,
    'name': 'fake_name',
    'description': 'fake_description',
    'status': 'creating',
    'consistencygroup_id': fake.CONSISTENCY_GROUP_ID,
}

fake_group = {
    'id': fake.GROUP_ID,
    'user_id': fake.USER_ID,
    'project_id': fake.PROJECT_ID,
    'host': 'fake_host',
    'availability_zone': 'fake_az',
    'name': 'fake_name',
    'description': 'fake_description',
    'group_type_id': fake.GROUP_TYPE_ID,
    'status': fields.GroupStatus.CREATING,
}


class TestConsistencyGroup(test_objects.BaseObjectsTestCase):

    @mock.patch('cinder.db.sqlalchemy.api.consistencygroup_get',
                return_value=fake_consistencygroup)
    def test_get_by_id(self, consistencygroup_get):
        consistencygroup = objects.ConsistencyGroup.get_by_id(
            self.context, fake.CONSISTENCY_GROUP_ID)
        self._compare(self, fake_consistencygroup, consistencygroup)
        consistencygroup_get.assert_called_once_with(
            self.context, fake.CONSISTENCY_GROUP_ID)

    @mock.patch('cinder.db.sqlalchemy.api.model_query')
    def test_get_by_id_no_existing_id(self, model_query):
        model_query().filter_by().first.return_value = None
        self.assertRaises(exception.ConsistencyGroupNotFound,
                          objects.ConsistencyGroup.get_by_id, self.context,
                          123)

    @mock.patch('cinder.db.consistencygroup_create',
                return_value=fake_consistencygroup)
    def test_create(self, consistencygroup_create):
        fake_cg = fake_consistencygroup.copy()
        del fake_cg['id']
        consistencygroup = objects.ConsistencyGroup(context=self.context,
                                                    **fake_cg)
        consistencygroup.create()
        self._compare(self, fake_consistencygroup, consistencygroup)

    @mock.patch('cinder.db.group_create',
                return_value=fake_group)
    def test_create_from_group(self, group_create):
        fake_grp = fake_group.copy()
        del fake_grp['id']
        group = objects.Group(context=self.context,
                              **fake_grp)
        group.create()
        volumes_objs = [objects.Volume(context=self.context, id=i)
                        for i in [fake.VOLUME_ID, fake.VOLUME2_ID,
                                  fake.VOLUME3_ID]]
        volumes = objects.VolumeList(objects=volumes_objs)
        group.volumes = volumes
        consistencygroup = objects.ConsistencyGroup()
        consistencygroup.from_group(group)
        self.assertEqual(group.id, consistencygroup.id)
        self.assertEqual(group.name, consistencygroup.name)

    def test_create_with_id_except_exception(self, ):
        consistencygroup = objects.ConsistencyGroup(
            context=self.context, **{'id': fake.CONSISTENCY_GROUP_ID})
        self.assertRaises(exception.ObjectActionError, consistencygroup.create)

    @mock.patch('cinder.db.consistencygroup_update')
    def test_save(self, consistencygroup_update):
        consistencygroup = objects.ConsistencyGroup._from_db_object(
            self.context, objects.ConsistencyGroup(), fake_consistencygroup)
        consistencygroup.status = fields.ConsistencyGroupStatus.AVAILABLE
        consistencygroup.save()
        consistencygroup_update.assert_called_once_with(
            self.context,
            consistencygroup.id,
            {'status': fields.ConsistencyGroupStatus.AVAILABLE})

    def test_save_with_cgsnapshots(self):
        consistencygroup = objects.ConsistencyGroup._from_db_object(
            self.context, objects.ConsistencyGroup(), fake_consistencygroup)
        cgsnapshots_objs = [objects.CGSnapshot(context=self.context, id=i)
                            for i in [fake.CGSNAPSHOT_ID, fake.CGSNAPSHOT2_ID,
                                      fake.CGSNAPSHOT3_ID]]
        cgsnapshots = objects.CGSnapshotList(objects=cgsnapshots_objs)
        consistencygroup.name = 'foobar'
        consistencygroup.cgsnapshots = cgsnapshots
        self.assertEqual({'name': 'foobar',
                          'cgsnapshots': cgsnapshots},
                         consistencygroup.obj_get_changes())
        self.assertRaises(exception.ObjectActionError, consistencygroup.save)

    def test_save_with_volumes(self):
        consistencygroup = objects.ConsistencyGroup._from_db_object(
            self.context, objects.ConsistencyGroup(), fake_consistencygroup)
        volumes_objs = [objects.Volume(context=self.context, id=i)
                        for i in [fake.VOLUME_ID, fake.VOLUME2_ID,
                                  fake.VOLUME3_ID]]
        volumes = objects.VolumeList(objects=volumes_objs)
        consistencygroup.name = 'foobar'
        consistencygroup.volumes = volumes
        self.assertEqual({'name': 'foobar',
                          'volumes': volumes},
                         consistencygroup.obj_get_changes())
        self.assertRaises(exception.ObjectActionError, consistencygroup.save)

    @mock.patch('cinder.objects.cgsnapshot.CGSnapshotList.get_all_by_group')
    @mock.patch('cinder.objects.volume.VolumeList.get_all_by_group')
    def test_obj_load_attr(self, mock_vol_get_all_by_group,
                           mock_cgsnap_get_all_by_group):
        consistencygroup = objects.ConsistencyGroup._from_db_object(
            self.context, objects.ConsistencyGroup(), fake_consistencygroup)
        # Test cgsnapshots lazy-loaded field
        cgsnapshots_objs = [objects.CGSnapshot(context=self.context, id=i)
                            for i in [fake.CGSNAPSHOT_ID, fake.CGSNAPSHOT2_ID,
                                      fake.CGSNAPSHOT3_ID]]
        cgsnapshots = objects.CGSnapshotList(context=self.context,
                                             objects=cgsnapshots_objs)
        mock_cgsnap_get_all_by_group.return_value = cgsnapshots
        self.assertEqual(cgsnapshots, consistencygroup.cgsnapshots)
        mock_cgsnap_get_all_by_group.assert_called_once_with(
            self.context, consistencygroup.id)

        # Test volumes lazy-loaded field
        volume_objs = [objects.Volume(context=self.context, id=i)
                       for i in [fake.VOLUME_ID, fake.VOLUME2_ID,
                                 fake.VOLUME3_ID]]
        volumes = objects.VolumeList(context=self.context, objects=volume_objs)
        mock_vol_get_all_by_group.return_value = volumes
        self.assertEqual(volumes, consistencygroup.volumes)
        mock_vol_get_all_by_group.assert_called_once_with(self.context,
                                                          consistencygroup.id)

    @mock.patch('oslo_utils.timeutils.utcnow', return_value=timeutils.utcnow())
    @mock.patch('cinder.db.sqlalchemy.api.consistencygroup_destroy')
    def test_destroy(self, consistencygroup_destroy, utcnow_mock):
        consistencygroup_destroy.return_value = {
            'status': fields.ConsistencyGroupStatus.DELETED,
            'deleted': True,
            'deleted_at': utcnow_mock.return_value}
        consistencygroup = objects.ConsistencyGroup(
            context=self.context, id=fake.CONSISTENCY_GROUP_ID)
        consistencygroup.destroy()
        self.assertTrue(consistencygroup_destroy.called)
        admin_context = consistencygroup_destroy.call_args[0][0]
        self.assertTrue(admin_context.is_admin)
        self.assertTrue(consistencygroup.deleted)
        self.assertEqual(fields.ConsistencyGroupStatus.DELETED,
                         consistencygroup.status)
        self.assertEqual(utcnow_mock.return_value.replace(tzinfo=pytz.UTC),
                         consistencygroup.deleted_at)

    @mock.patch('cinder.db.sqlalchemy.api.consistencygroup_get')
    def test_refresh(self, consistencygroup_get):
        db_cg1 = fake_consistencygroup.copy()
        db_cg2 = db_cg1.copy()
        db_cg2['description'] = 'foobar'

        # On the second consistencygroup_get, return the ConsistencyGroup with
        # an updated description
        consistencygroup_get.side_effect = [db_cg1, db_cg2]
        cg = objects.ConsistencyGroup.get_by_id(self.context,
                                                fake.CONSISTENCY_GROUP_ID)
        self._compare(self, db_cg1, cg)

        # description was updated, so a ConsistencyGroup refresh should have a
        # new value for that field
        cg.refresh()
        self._compare(self, db_cg2, cg)
        if six.PY3:
            call_bool = mock.call.__bool__()
        else:
            call_bool = mock.call.__nonzero__()
        consistencygroup_get.assert_has_calls([
            mock.call(
                self.context,
                fake.CONSISTENCY_GROUP_ID),
            call_bool,
            mock.call(
                self.context,
                fake.CONSISTENCY_GROUP_ID)])

    def test_from_db_object_with_all_expected_attributes(self):
        expected_attrs = ['volumes', 'cgsnapshots']
        db_volumes = [fake_volume.fake_db_volume(admin_metadata={},
                                                 volume_metadata={})]
        db_cgsnaps = [fake_cgsnapshot.copy()]
        db_cg = fake_consistencygroup.copy()
        db_cg['volumes'] = db_volumes
        db_cg['cgsnapshots'] = db_cgsnaps
        cg = objects.ConsistencyGroup._from_db_object(
            self.context, objects.ConsistencyGroup(), db_cg, expected_attrs)
        self.assertEqual(len(db_volumes), len(cg.volumes))
        self._compare(self, db_volumes[0], cg.volumes[0])
        self.assertEqual(len(db_cgsnaps), len(cg.cgsnapshots))
        self._compare(self, db_cgsnaps[0], cg.cgsnapshots[0])


class TestConsistencyGroupList(test_objects.BaseObjectsTestCase):
    @mock.patch('cinder.db.consistencygroup_get_all',
                return_value=[fake_consistencygroup])
    def test_get_all(self, consistencygroup_get_all):
        consistencygroups = objects.ConsistencyGroupList.get_all(self.context)
        self.assertEqual(1, len(consistencygroups))
        TestConsistencyGroup._compare(self, fake_consistencygroup,
                                      consistencygroups[0])

    @mock.patch('cinder.db.consistencygroup_get_all_by_project',
                return_value=[fake_consistencygroup])
    def test_get_all_by_project(self, consistencygroup_get_all_by_project):
        consistencygroups = objects.ConsistencyGroupList.get_all_by_project(
            self.context, self.project_id)
        self.assertEqual(1, len(consistencygroups))
        TestConsistencyGroup._compare(self, fake_consistencygroup,
                                      consistencygroups[0])

    @mock.patch('cinder.db.consistencygroup_get_all',
                return_value=[fake_consistencygroup])
    def test_get_all_with_pagination(self, consistencygroup_get_all):
        consistencygroups = objects.ConsistencyGroupList.get_all(
            self.context, filters={'id': 'fake'}, marker=None, limit=1,
            offset=None, sort_keys='id', sort_dirs='asc')
        self.assertEqual(1, len(consistencygroups))
        consistencygroup_get_all.assert_called_once_with(
            self.context, filters={'id': 'fake'}, marker=None, limit=1,
            offset=None, sort_keys='id', sort_dirs='asc')
        TestConsistencyGroup._compare(self, fake_consistencygroup,
                                      consistencygroups[0])

    @mock.patch('cinder.db.consistencygroup_get_all_by_project',
                return_value=[fake_consistencygroup])
    def test_get_all_by_project_with_pagination(
            self, consistencygroup_get_all_by_project):
        consistencygroups = objects.ConsistencyGroupList.get_all_by_project(
            self.context, self.project_id, filters={'id': 'fake'}, marker=None,
            limit=1, offset=None, sort_keys='id', sort_dirs='asc')
        self.assertEqual(1, len(consistencygroups))
        consistencygroup_get_all_by_project.assert_called_once_with(
            self.context, self.project_id, filters={'id': 'fake'}, marker=None,
            limit=1, offset=None, sort_keys='id', sort_dirs='asc')
        TestConsistencyGroup._compare(self, fake_consistencygroup,
                                      consistencygroups[0])

    @mock.patch('cinder.db.consistencygroup_include_in_cluster')
    def test_include_in_cluster(self, include_mock):
        filters = {'host': mock.sentinel.host,
                   'cluster_name': mock.sentinel.cluster_name}
        cluster = 'new_cluster'
        objects.ConsistencyGroupList.include_in_cluster(self.context, cluster,
                                                        **filters)
        include_mock.assert_called_once_with(self.context, cluster, True,
                                             **filters)

    @mock.patch('cinder.db.consistencygroup_include_in_cluster')
    def test_include_in_cluster_specify_partial(self, include_mock):
        filters = {'host': mock.sentinel.host,
                   'cluster_name': mock.sentinel.cluster_name}
        cluster = 'new_cluster'
        objects.ConsistencyGroupList.include_in_cluster(
            self.context, cluster, mock.sentinel.partial_rename, **filters)
        include_mock.assert_called_once_with(
            self.context, cluster, mock.sentinel.partial_rename, **filters)



# Copyright (c) 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Scheduler host filters
"""

from cinder.scheduler import base_filter


class BaseBackendFilter(base_filter.BaseFilter):
    """Base class for host filters."""
    def _filter_one(self, obj, filter_properties):
        """Return True if the object passes the filter, otherwise False."""
        # For backward compatibility with out of tree filters
        passes_method = getattr(self, 'host_passes', self.backend_passes)
        return passes_method(obj, filter_properties)

    def backend_passes(self, host_state, filter_properties):
        """Return True if the HostState passes the filter, otherwise False.

        Override this in a subclass.
        """
        raise NotImplementedError()


class BackendFilterHandler(base_filter.BaseFilterHandler):
    def __init__(self, namespace):
        super(BackendFilterHandler, self).__init__(BaseHostFilter, namespace)


# NOTE(geguileo): For backward compatibility with external filters that
# inherit from these classes
BaseHostFilter = BaseBackendFilter
HostFilterHandler = BackendFilterHandler



# Copyright (c) 2015 Infortrend Technology, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Fibre Channel Driver for Infortrend Eonstor based on CLI.
"""


from oslo_log import log as logging

from cinder import interface
from cinder.volume import driver
from cinder.volume.drivers.infortrend.raidcmd_cli import common_cli

LOG = logging.getLogger(__name__)


@interface.volumedriver
class InfortrendCLIFCDriver(driver.FibreChannelDriver):

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = "Infortrend_Storage_CI"
    VERSION = common_cli.InfortrendCommon.VERSION

    def __init__(self, *args, **kwargs):
        super(InfortrendCLIFCDriver, self).__init__(*args, **kwargs)
        self.common = common_cli.InfortrendCommon(
            'FC', configuration=self.configuration)
        self.VERSION = self.common.VERSION

    @staticmethod
    def get_driver_options():
        """Return the oslo_config options specific to the driver."""
        return common_cli.infortrend_opts

    def do_setup(self, context):
        """Any initialization the volume driver does while starting.

            note: This runs before check_for_setup_error
        """

        LOG.debug('do_setup start')
        self.common.do_setup()

    def check_for_setup_error(self):
        LOG.debug('check_for_setup_error start')
        self.common.check_for_setup_error()

    def create_volume(self, volume):
        """Creates a volume.

        Can optionally return a Dictionary of changes
        to the volume object to be persisted.
        """
        LOG.debug('create_volume volume id=%(volume_id)s', {
            'volume_id': volume['id']})
        return self.common.create_volume(volume)

    def create_volume_from_snapshot(self, volume, snapshot):
        """Creates a volume from a snapshot."""
        LOG.debug(
            'create_volume_from_snapshot volume id=%(volume_id)s '
            'snapshot id=%(snapshot_id)s', {
                'volume_id': volume['id'], 'snapshot_id': snapshot['id']})
        return self.common.create_volume_from_snapshot(volume, snapshot)

    def create_cloned_volume(self, volume, src_vref):
        """Creates a clone of the specified volume."""
        LOG.debug(
            'create_cloned_volume volume id=%(volume_id)s '
            'src_vref provider_location=%(provider_location)s', {
                'volume_id': volume['id'],
                'provider_location': src_vref['provider_location']})
        return self.common.create_cloned_volume(volume, src_vref)

    def extend_volume(self, volume, new_size):
        """Extend a volume."""
        LOG.debug(
            'extend_volume volume id=%(volume_id)s new size=%(size)s', {
                'volume_id': volume['id'], 'size': new_size})
        self.common.extend_volume(volume, new_size)

    def delete_volume(self, volume):
        """Deletes a volume."""
        LOG.debug('delete_volume volume id=%(volume_id)s', {
            'volume_id': volume['id']})
        return self.common.delete_volume(volume)

    def migrate_volume(self, ctxt, volume, host):
        """Migrate the volume to the specified host.

        Returns a boolean indicating whether the migration occurred, as well as
        model_update.

        :param ctxt: Context
        :param volume: A dictionary describing the volume to migrate
        :param host: A dictionary describing the host to migrate to, where
                     host['host'] is its name, and host['capabilities'] is a
                     dictionary of its reported capabilities.
        """
        LOG.debug('migrate_volume volume id=%(volume_id)s host=%(host)s', {
            'volume_id': volume['id'], 'host': host['host']})
        return self.common.migrate_volume(volume, host)

    def create_snapshot(self, snapshot):
        """Creates a snapshot."""
        LOG.debug(
            'create_snapshot snapshot id=%(snapshot_id)s '
            'volume id=%(volume_id)s', {
                'snapshot_id': snapshot['id'],
                'volume_id': snapshot['volume_id']})
        return self.common.create_snapshot(snapshot)

    def delete_snapshot(self, snapshot):
        """Deletes a snapshot."""
        LOG.debug(
            'delete_snapshot snapshot id=%(snapshot_id)s '
            'volume id=%(volume_id)s', {
                'snapshot_id': snapshot['id'],
                'volume_id': snapshot['volume_id']})
        self.common.delete_snapshot(snapshot)

    def ensure_export(self, context, volume):
        """Synchronously recreates an export for a volume."""
        pass

    def create_export(self, context, volume, connector):
        """Exports the volume.

        Can optionally return a Dictionary of changes
        to the volume object to be persisted.
        """
        LOG.debug(
            'create_export volume provider_location=%(provider_location)s', {
                'provider_location': volume['provider_location']})
        return self.common.create_export(context, volume)

    def remove_export(self, context, volume):
        """Removes an export for a volume."""
        pass

    def initialize_connection(self, volume, connector):
        """Initializes the connection and returns connection information.

        Assign any created volume to a compute node/host so that it can be
        used from that host.

        The  driver returns a driver_volume_type of 'fibre_channel'.
        The target_wwn can be a single entry or a list of wwns that
        correspond to the list of remote wwn(s) that will export the volume.
        The initiator_target_map is a map that represents the remote wwn(s)
        and a list of wwns which are visible to the remote wwn(s).
        Example return values:

            {
                'driver_volume_type': 'fibre_channel'
                'data': {
                    'target_discovered': True,
                    'target_lun': 1,
                    'target_wwn': '1234567890123',
                    'initiator_target_map': {
                        '1122334455667788': ['1234567890123']
                    }
                }
            }

            or

            {
                'driver_volume_type': 'fibre_channel'
                'data': {
                    'target_discovered': True,
                    'target_lun': 1,
                    'target_wwn': ['1234567890123', '0987654321321'],
                    'initiator_target_map': {
                        '1122334455667788': ['1234567890123',
                                             '0987654321321']
                    }
                }
            }
        """
        LOG.debug(
            'initialize_connection volume id=%(volume_id)s '
            'connector initiator=%(initiator)s', {
                'volume_id': volume['id'],
                'initiator': connector['initiator']})
        return self.common.initialize_connection(volume, connector)

    def terminate_connection(self, volume, connector, **kwargs):
        """Disallow connection from connector."""
        LOG.debug('terminate_connection volume id=%(volume_id)s', {
            'volume_id': volume['id']})
        return self.common.terminate_connection(volume, connector)

    def get_volume_stats(self, refresh=False):
        """Get volume stats.

        If 'refresh' is True, run update the stats first.
        """
        LOG.debug('get_volume_stats refresh=%(refresh)s', {
            'refresh': refresh})
        return self.common.get_volume_stats(refresh)

    def manage_existing(self, volume, existing_ref):
        """Manage an existing lun in the array.

        The lun should be in a manageable pool backend, otherwise
        error would return.
        Rename the backend storage object so that it matches the,
        volume['name'] which is how drivers traditionally map between a
        cinder volume and the associated backend storage object.

        :param existing_ref: Driver-specific information used to identify
                             a volume
        """
        LOG.debug(
            'manage_existing volume: %(volume)s '
            'existing_ref source: %(source)s', {
                'volume': volume,
                'source': existing_ref})
        return self.common.manage_existing(volume, existing_ref)

    def unmanage(self, volume):
        """Removes the specified volume from Cinder management.

        Does not delete the underlying backend storage object.

        :param volume: Cinder volume to unmanage
        """
        LOG.debug('unmanage volume id=%(volume_id)s', {
            'volume_id': volume['id']})
        self.common.unmanage(volume)

    def manage_existing_get_size(self, volume, existing_ref):
        """Return size of volume to be managed by manage_existing.

        When calculating the size, round up to the next GB.
        """
        LOG.debug(
            'manage_existing_get_size volume: %(volume)s '
            'existing_ref source: %(source)s', {
                'volume': volume,
                'source': existing_ref})
        return self.common.manage_existing_get_size(volume, existing_ref)

    def retype(self, ctxt, volume, new_type, diff, host):
        """Convert the volume to be of the new type.

        :param ctxt: Context
        :param volume: A dictionary describing the volume to migrate
        :param new_type: A dictionary describing the volume type to convert to
        :param diff: A dictionary with the difference between the two types
        :param host: A dictionary describing the host to migrate to, where
                     host['host'] is its name, and host['capabilities'] is a
                     dictionary of its reported capabilities.
        """
        LOG.debug(
            'retype volume id=%(volume_id)s new_type id=%(type_id)s', {
                'volume_id': volume['id'], 'type_id': new_type['id']})
        return self.common.retype(ctxt, volume, new_type, diff, host)

    def update_migrated_volume(self, ctxt, volume, new_volume,
                               original_volume_status):
        """Return model update for migrated volume.

        :param volume: The original volume that was migrated to this backend
        :param new_volume: The migration volume object that was created on
                           this backend as part of the migration process
        :param original_volume_status: The status of the original volume
        :returns: model_update to update DB with any needed changes
        """
        LOG.debug(
            'update migrated volume original volume id= %(volume_id)s '
            'new volume id=%(new_volume_id)s', {
                'volume_id': volume['id'], 'new_volume_id': new_volume['id']})
        return self.common.update_migrated_volume(ctxt, volume, new_volume,
                                                  original_volume_status)

    def get_manageable_volumes(self, cinder_volumes, marker, limit, offset,
                               sort_keys, sort_dirs):
        """List volumes on the backend available for management by Cinder."""

        LOG.debug(
            'get_manageable_volumes CALLED '
            'cinder_volumes: %(volume)s, '
            'marker: %(mkr)s, '
            'limit: %(lmt)s, '
            'offset: %(_offset)s, '
            'sort_keys: %(s_key)s, '
            'sort_dirs: %(sort_dir)s', {
                'volume': cinder_volumes,
                'mkr': marker,
                'lmt': limit,
                '_offset': offset,
                's_key': sort_keys,
                'sort_dir': sort_dirs
            }
        )
        return self.common.get_manageable_volumes(cinder_volumes, marker,
                                                  limit, offset, sort_keys,
                                                  sort_dirs)

    def manage_existing_snapshot(self, snapshot, existing_ref):
        """Brings an existing backend storage object under Cinder management.

        :param snapshot:     Cinder volume snapshot to manage
        :param existing_ref: Driver-specific information used to identify a
                             volume snapshot
        """

        LOG.debug(
            'manage_existing_snapshot CALLED '
            'snapshot: %(si)s, '
            'existing_ref: %(ref)s', {
                'si': snapshot, 'ref': existing_ref
            }
        )
        return self.common.manage_existing_snapshot(snapshot, existing_ref)

    def manage_existing_snapshot_get_size(self, snapshot, existing_ref):
        """Return size of snapshot to be managed by manage_existing.

        :param snapshot:     Cinder volume snapshot to manage
        :param existing_ref: Driver-specific information used to identify a
                             volume snapshot
        :returns size:       Volume snapshot size in GiB (integer)
        """

        LOG.debug(
            'manage_existing_snapshot_get_size CALLED '
            'snapshot: %(si)s, '
            'existing_ref: %(ref)s', {
                'si': snapshot, 'ref': existing_ref
            }
        )
        return self.common.manage_existing_snapshot_get_size(snapshot,
                                                             existing_ref)

    def get_manageable_snapshots(self, cinder_snapshots, marker, limit, offset,
                                 sort_keys, sort_dirs):
        """List snapshots on the backend available for management by Cinder."""

        LOG.debug(
            'get_manageable_volumes CALLED '
            'cinder_snapshots: %(volume)s, '
            'marker: %(mkr)s, '
            'limit: %(lmt)s, '
            'offset: %(_offset)s, '
            'sort_keys: %(s_key)s, '
            'sort_dirs: %(sort_dir)s', {
                'volume': cinder_snapshots,
                'mkr': marker,
                'lmt': limit,
                '_offset': offset,
                's_key': sort_keys,
                'sort_dir': sort_dirs
            }
        )
        return self.common.get_manageable_snapshots(cinder_snapshots, marker,
                                                    limit, offset, sort_keys,
                                                    sort_dirs)

    def unmanage_snapshot(self, snapshot):
        """Removes the specified snapshot from Cinder management.

        Does not delete the underlying backend storage object.

        For most drivers, this will not need to do anything. However, some
        drivers might use this call as an opportunity to clean up any
        Cinder-specific configuration that they have associated with the
        backend storage object.

        :param snapshot: Cinder volume snapshot to unmanage
        """
        LOG.debug(
            'manage_existing_snapshot_get_size CALLED '
            'snapshot: %(si)s', {
                'si': snapshot
            }
        )
        return self.common.unmanage_snapshot(snapshot)



# Copyright 2016 EMC Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
import uuid

import ddt
from oslo_utils import strutils
from oslo_utils import timeutils
import six
import webob

from cinder.api import microversions as mv
from cinder.api.v3 import group_specs as v3_group_specs
from cinder.api.v3 import group_types as v3_group_types
from cinder.api.v3.views import group_types as views_types
from cinder import context
from cinder import exception
from cinder.tests.unit.api import fakes
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import test
from cinder.volume import group_types

IN_USE_GROUP_TYPE = fake.GROUP_TYPE3_ID


def stub_group_type(id):
    specs = {
        "key1": "value1",
        "key2": "value2",
        "key3": "value3",
        "key4": "value4",
        "key5": "value5"
    }
    return dict(
        id=id,
        name='group_type_%s' % six.text_type(id),
        description='group_type_desc_%s' % six.text_type(id),
        group_specs=specs,
    )


def return_group_types_get_all_types(context, filters=None, marker=None,
                                     limit=None, sort_keys=None,
                                     sort_dirs=None, offset=None,
                                     list_result=False):
    result = dict(group_type_1=stub_group_type(1),
                  group_type_2=stub_group_type(2),
                  group_type_3=stub_group_type(3)
                  )
    if list_result:
        return list(result.values())
    return result


def return_empty_group_types_get_all_types(context, filters=None, marker=None,
                                           limit=None, sort_keys=None,
                                           sort_dirs=None, offset=None,
                                           list_result=False):
    if list_result:
        return []
    return {}


def return_group_types_get_group_type(context, id):
    if id == fake.WILL_NOT_BE_FOUND_ID:
        raise exception.GroupTypeNotFound(group_type_id=id)
    return stub_group_type(id)


def return_group_types_get_default():
    return stub_group_type(1)


def return_group_types_get_default_not_found():
    return {}


def return_group_types_with_groups_destroy(context, id):
    if id == IN_USE_GROUP_TYPE:
        raise exception.GroupTypeInUse(group_type_id=id)


@ddt.ddt
class GroupTypesApiTest(test.TestCase):

    def _create_group_type(self, group_type_name, group_specs=None,
                           is_public=True, projects=None):
        return group_types.create(self.ctxt, group_type_name, group_specs,
                                  is_public, projects).get('id')

    def setUp(self):
        super(GroupTypesApiTest, self).setUp()
        self.controller = v3_group_types.GroupTypesController()
        self.specs_controller = v3_group_specs.GroupTypeSpecsController()
        self.ctxt = context.RequestContext(user_id=fake.USER_ID,
                                           project_id=fake.PROJECT_ID,
                                           is_admin=True)
        self.user_ctxt = context.RequestContext(user_id=fake.USER2_ID,
                                                project_id=fake.PROJECT2_ID,
                                                is_admin=False)
        self.type_id1 = self._create_group_type('group_type1',
                                                {'key1': 'value1'})
        self.type_id2 = self._create_group_type('group_type2',
                                                {'key2': 'value2'})
        self.type_id3 = self._create_group_type('group_type3',
                                                {'key3': 'value3'}, False,
                                                [fake.PROJECT_ID])
        self.type_id0 = group_types.get_default_cgsnapshot_type()['id']

    @ddt.data('0', 'f', 'false', 'off', 'n', 'no', '1', 't', 'true', 'on',
              'y', 'yes')
    @mock.patch.object(group_types, "get_group_type_by_name")
    @mock.patch.object(group_types, "create")
    @mock.patch("cinder.api.openstack.wsgi.Request.cache_resource")
    @mock.patch("cinder.api.views.types.ViewBuilder.show")
    def test_create_group_type_with_valid_is_public_in_string(
            self, is_public, mock_show, mock_cache_resource,
            mock_create, mock_get):
        boolean_is_public = strutils.bool_from_string(is_public)
        req = fakes.HTTPRequest.blank('/v3/%s/types' % fake.PROJECT_ID,
                                      version=mv.GROUP_TYPE)
        req.environ['cinder.context'] = self.ctxt

        body = {"group_type": {"is_public": is_public, "name": "group_type1",
                               "description": None}}
        self.controller.create(req, body=body)
        mock_create.assert_called_once_with(
            self.ctxt, 'group_type1', {},
            boolean_is_public, description=None)

    @mock.patch.object(group_types, "get_group_type_by_name")
    @mock.patch.object(group_types, "create")
    @mock.patch("cinder.api.openstack.wsgi.Request.cache_resource")
    @mock.patch("cinder.api.views.types.ViewBuilder.show")
    def test_create_group_type_with_group_specs_null(
            self, mock_show, mock_cache_resource,
            mock_create, mock_get):

        req = fakes.HTTPRequest.blank('/v3/%s/types' % fake.PROJECT_ID,
                                      version=mv.GROUP_TYPE)
        req.environ['cinder.context'] = self.ctxt

        body = {"group_type": {"name": "group_type1",
                               "group_specs": None}}
        self.controller.create(req, body=body)
        mock_create.assert_called_once_with(
            self.ctxt, 'group_type1', None, True, description=None)

    @ddt.data(fake.GROUP_TYPE_ID, IN_USE_GROUP_TYPE)
    def test_group_type_destroy(self, grp_type_id):
        grp_type = {'id': grp_type_id, 'name': 'grp' + grp_type_id}
        self.mock_object(group_types, 'get_group_type',
                         return_value=grp_type)
        self.mock_object(group_types, 'destroy',
                         return_group_types_with_groups_destroy)
        mock_notify_info = self.mock_object(
            v3_group_types.GroupTypesController,
            '_notify_group_type_info')
        mock_notify_error = self.mock_object(
            v3_group_types.GroupTypesController,
            '_notify_group_type_error')
        req = fakes.HTTPRequest.blank('/v3/%s/group_types/%s' % (
            fake.PROJECT_ID, grp_type_id),
            version=mv.GROUP_TYPE)
        req.environ['cinder.context'] = self.ctxt
        if grp_type_id == IN_USE_GROUP_TYPE:
            self.assertRaises(webob.exc.HTTPBadRequest,
                              self.controller.delete,
                              req, grp_type_id)
            mock_notify_error.assert_called_once_with(
                self.ctxt, 'group_type.delete', mock.ANY,
                group_type=grp_type)
        else:
            self.controller.delete(req, grp_type_id)
            mock_notify_info.assert_called_once_with(
                self.ctxt, 'group_type.delete', grp_type)

    def test_group_types_index(self):
        self.mock_object(group_types, 'get_all_group_types',
                         return_group_types_get_all_types)

        req = fakes.HTTPRequest.blank('/v3/%s/group_types' % fake.PROJECT_ID,
                                      use_admin_context=True,
                                      version=mv.GROUP_TYPE)
        res_dict = self.controller.index(req)

        self.assertEqual(3, len(res_dict['group_types']))

        expected_names = ['group_type_1', 'group_type_2', 'group_type_3']
        actual_names = map(lambda e: e['name'], res_dict['group_types'])
        self.assertEqual(set(expected_names), set(actual_names))
        for entry in res_dict['group_types']:
            self.assertEqual('value1', entry['group_specs']['key1'])

    def test_group_types_index_no_data(self):
        self.mock_object(group_types, 'get_all_group_types',
                         return_empty_group_types_get_all_types)

        req = fakes.HTTPRequest.blank('/v3/%s/group_types' % fake.PROJECT_ID,
                                      version=mv.GROUP_TYPE)
        res_dict = self.controller.index(req)

        self.assertEqual(0, len(res_dict['group_types']))

    def test_group_types_index_with_limit(self):
        req = fakes.HTTPRequest.blank('/v3/%s/group_types?limit=1' %
                                      fake.PROJECT_ID,
                                      version=mv.GROUP_TYPE)
        req.environ['cinder.context'] = self.ctxt
        res = self.controller.index(req)

        self.assertEqual(1, len(res['group_types']))
        self.assertEqual(self.type_id3, res['group_types'][0]['id'])

        expect_next_link = ('http://localhost/v3/%s/group_types?limit=1'
                            '&marker=%s' %
                            (fake.PROJECT_ID, res['group_types'][0]['id']))
        self.assertEqual(expect_next_link, res['group_type_links'][0]['href'])

    def test_group_types_index_with_offset(self):
        req = fakes.HTTPRequest.blank(
            '/v3/%s/group_types?offset=1' % fake.PROJECT_ID,
            version=mv.GROUP_TYPE)
        req.environ['cinder.context'] = self.ctxt
        res = self.controller.index(req)

        self.assertEqual(3, len(res['group_types']))

    def test_group_types_index_with_offset_out_of_range(self):
        url = '/v3/%s/group_types?offset=424366766556787' % fake.PROJECT_ID
        req = fakes.HTTPRequest.blank(url, version=mv.GROUP_TYPE)
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.index, req)

    def test_group_types_index_with_limit_and_offset(self):
        req = fakes.HTTPRequest.blank(
            '/v3/%s/group_types?limit=2&offset=1' % fake.PROJECT_ID,
            version=mv.GROUP_TYPE)
        req.environ['cinder.context'] = self.ctxt
        res = self.controller.index(req)

        self.assertEqual(2, len(res['group_types']))
        self.assertEqual(self.type_id2, res['group_types'][0]['id'])
        self.assertEqual(self.type_id1, res['group_types'][1]['id'])

    def test_group_types_index_with_limit_and_marker(self):
        req = fakes.HTTPRequest.blank('/v3/%s/group_types?limit=1'
                                      '&marker=%s' %
                                      (fake.PROJECT_ID,
                                       self.type_id2),
                                      version=mv.GROUP_TYPE)
        req.environ['cinder.context'] = self.ctxt
        res = self.controller.index(req)

        self.assertEqual(1, len(res['group_types']))
        self.assertEqual(self.type_id1, res['group_types'][0]['id'])

    def test_group_types_index_with_valid_filter(self):
        req = fakes.HTTPRequest.blank(
            '/v3/%s/group_types?is_public=True' % fake.PROJECT_ID,
            version=mv.GROUP_TYPE)
        req.environ['cinder.context'] = self.ctxt
        res = self.controller.index(req)

        self.assertEqual(4, len(res['group_types']))
        self.assertEqual(self.type_id3, res['group_types'][0]['id'])
        self.assertEqual(self.type_id2, res['group_types'][1]['id'])
        self.assertEqual(self.type_id1, res['group_types'][2]['id'])
        self.assertEqual(self.type_id0, res['group_types'][3]['id'])

    def test_group_types_index_with_invalid_filter(self):
        req = fakes.HTTPRequest.blank(
            '/v3/%s/group_types?id=%s' % (fake.PROJECT_ID, self.type_id1),
            version=mv.GROUP_TYPE)
        req.environ['cinder.context'] = self.ctxt
        res = self.controller.index(req)

        self.assertEqual(4, len(res['group_types']))

    def test_group_types_index_with_sort_keys(self):
        req = fakes.HTTPRequest.blank('/v3/%s/group_types?sort=id' %
                                      fake.PROJECT_ID,
                                      version=mv.GROUP_TYPE)
        req.environ['cinder.context'] = self.ctxt
        res = self.controller.index(req)
        expect_result = [self.type_id0, self.type_id1, self.type_id2,
                         self.type_id3]
        expect_result.sort(reverse=True)

        self.assertEqual(4, len(res['group_types']))
        self.assertEqual(expect_result[0], res['group_types'][0]['id'])
        self.assertEqual(expect_result[1], res['group_types'][1]['id'])
        self.assertEqual(expect_result[2], res['group_types'][2]['id'])
        self.assertEqual(expect_result[3], res['group_types'][3]['id'])

    def test_group_types_index_with_sort_and_limit(self):
        req = fakes.HTTPRequest.blank(
            '/v3/%s/group_types?sort=id&limit=2' % fake.PROJECT_ID,
            version=mv.GROUP_TYPE)
        req.environ['cinder.context'] = self.ctxt
        res = self.controller.index(req)
        expect_result = [self.type_id0, self.type_id1, self.type_id2,
                         self.type_id3]
        expect_result.sort(reverse=True)

        self.assertEqual(2, len(res['group_types']))
        self.assertEqual(expect_result[0], res['group_types'][0]['id'])
        self.assertEqual(expect_result[1], res['group_types'][1]['id'])

    def test_group_types_index_with_sort_keys_and_sort_dirs(self):
        req = fakes.HTTPRequest.blank(
            '/v3/%s/group_types?sort=id:asc' % fake.PROJECT_ID,
            version=mv.GROUP_TYPE)
        req.environ['cinder.context'] = self.ctxt
        res = self.controller.index(req)
        expect_result = [self.type_id0, self.type_id1, self.type_id2,
                         self.type_id3]
        expect_result.sort()

        self.assertEqual(4, len(res['group_types']))
        self.assertEqual(expect_result[0], res['group_types'][0]['id'])
        self.assertEqual(expect_result[1], res['group_types'][1]['id'])
        self.assertEqual(expect_result[2], res['group_types'][2]['id'])
        self.assertEqual(expect_result[3], res['group_types'][3]['id'])

    @ddt.data('0', 'f', 'false', 'off', 'n', 'no', '1', 't', 'true', 'on',
              'y', 'yes')
    @mock.patch.object(group_types, "get_group_type")
    @mock.patch.object(group_types, "update")
    @mock.patch("cinder.api.openstack.wsgi.Request.cache_resource")
    @mock.patch("cinder.api.views.types.ViewBuilder.show")
    def test_update_group_type_with_valid_is_public_in_string(
            self, is_public, mock_show, mock_cache_resource,
            mock_update, mock_get):
        type_id = six.text_type(uuid.uuid4())
        req = fakes.HTTPRequest.blank(
            '/v3/%s/types/%s' % (fake.PROJECT_ID, type_id),
            version=mv.GROUP_TYPE)
        req.environ['cinder.context'] = self.ctxt
        boolean_is_public = strutils.bool_from_string(is_public)
        body = {"group_type": {"is_public": is_public, "name": "group_type1"}}
        self.controller.update(req, type_id, body=body)
        mock_update.assert_called_once_with(
            self.ctxt, type_id, 'group_type1', None,
            is_public=boolean_is_public)

    def test_update_group_type_with_name_null(self):
        req = fakes.HTTPRequest.blank(
            '/v3/%s/types/%s' % (fake.PROJECT_ID, fake.GROUP_TYPE_ID),
            version=mv.GROUP_TYPE)
        req.environ['cinder.context'] = self.ctxt
        body = {"group_type": {"name": None}}
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.update,
                          req, fake.GROUP_TYPE_ID, body=body)

    @ddt.data({"group_type": {"name": None,
                              "description": "description"}},
              {"group_type": {"name": "test",
                              "is_public": True}},
              {"group_type": {"description": None,
                              "is_public": True}})
    def test_update_group_type(self, body):
        req = fakes.HTTPRequest.blank(
            '/v3/%s/types/%s' % (fake.PROJECT_ID, fake.GROUP_TYPE_ID),
            version=mv.GROUP_TYPE)

        group_type_1 = group_types.create(self.ctxt, 'group_type')

        req.environ['cinder.context'] = self.ctxt
        res = self.controller.update(req, group_type_1.get('id'), body=body)

        expected_name = body['group_type'].get('name')
        if expected_name is not None:
            self.assertEqual(expected_name, res['group_type']['name'])

        expected_is_public = body['group_type'].get('is_public')
        if expected_is_public is not None:
            self.assertEqual(expected_is_public,
                             res['group_type']['is_public'])

        self.assertEqual(body['group_type'].get('description'),
                         res['group_type']['description'])

    def test_group_types_show(self):
        self.mock_object(group_types, 'get_group_type',
                         return_group_types_get_group_type)

        type_id = six.text_type(uuid.uuid4())
        req = fakes.HTTPRequest.blank('/v3/%s/group_types/' % fake.PROJECT_ID
                                      + type_id,
                                      version=mv.GROUP_TYPE)
        res_dict = self.controller.show(req, type_id)

        self.assertEqual(1, len(res_dict))
        self.assertEqual(type_id, res_dict['group_type']['id'])
        type_name = 'group_type_' + type_id
        self.assertEqual(type_name, res_dict['group_type']['name'])

    def test_group_types_show_pre_microversion(self):
        self.mock_object(group_types, 'get_group_type',
                         return_group_types_get_group_type)

        type_id = uuid.uuid4()
        req = fakes.HTTPRequest.blank(
            '/v3/%s/group_types/%s' % (fake.PROJECT_ID, type_id),
            version=mv.get_prior_version(mv.GROUP_TYPE))

        self.assertRaises(exception.VersionNotFoundForAPIMethod,
                          self.controller.show, req, type_id)

    def test_group_types_show_not_found(self):
        self.mock_object(group_types, 'get_group_type',
                         return_group_types_get_group_type)

        req = fakes.HTTPRequest.blank('/v3/%s/group_types/%s' %
                                      (fake.PROJECT_ID,
                                       fake.WILL_NOT_BE_FOUND_ID),
                                      version=mv.GROUP_TYPE)
        self.assertRaises(webob.exc.HTTPNotFound, self.controller.show,
                          req, fake.WILL_NOT_BE_FOUND_ID)

    def test_get_default(self):
        self.mock_object(group_types, 'get_default_group_type',
                         return_group_types_get_default)
        req = fakes.HTTPRequest.blank('/v3/%s/group_types/default' %
                                      fake.PROJECT_ID,
                                      version=mv.GROUP_TYPE)
        req.method = 'GET'
        res_dict = self.controller.show(req, 'default')
        self.assertEqual(1, len(res_dict))
        self.assertEqual('group_type_1', res_dict['group_type']['name'])
        self.assertEqual('group_type_desc_1',
                         res_dict['group_type']['description'])

    def test_get_default_not_found(self):
        self.mock_object(group_types, 'get_default_group_type',
                         return_group_types_get_default_not_found)
        req = fakes.HTTPRequest.blank('/v3/%s/group_types/default' %
                                      fake.PROJECT_ID,
                                      version=mv.GROUP_TYPE)
        req.method = 'GET'

        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.show, req, 'default')

    def test_view_builder_show(self):
        view_builder = views_types.ViewBuilder()

        now = timeutils.utcnow().isoformat()
        raw_group_type = dict(
            name='new_type',
            description='new_type_desc',
            is_public=True,
            deleted=False,
            created_at=now,
            updated_at=now,
            group_specs={},
            deleted_at=None,
            id=42,
        )

        request = fakes.HTTPRequest.blank("/v3",
                                          version=mv.GROUP_TYPE)
        output = view_builder.show(request, raw_group_type)

        self.assertIn('group_type', output)
        expected_group_type = dict(
            name='new_type',
            description='new_type_desc',
            is_public=True,
            id=42,
        )
        self.assertDictEqual(expected_group_type, output['group_type'])

    def test_view_builder_show_admin(self):
        view_builder = views_types.ViewBuilder()

        now = timeutils.utcnow().isoformat()
        raw_group_type = dict(
            name='new_type',
            description='new_type_desc',
            is_public=True,
            deleted=False,
            created_at=now,
            updated_at=now,
            group_specs={},
            deleted_at=None,
            id=42,
        )

        request = fakes.HTTPRequest.blank("/v3", use_admin_context=True,
                                          version=mv.GROUP_TYPE)
        output = view_builder.show(request, raw_group_type)

        self.assertIn('group_type', output)
        expected_group_type = dict(
            name='new_type',
            description='new_type_desc',
            is_public=True,
            group_specs={},
            id=42,
        )
        self.assertDictEqual(expected_group_type, output['group_type'])

    def __test_view_builder_show_qos_specs_id_policy(self):
        with mock.patch.object(context.RequestContext,
                               'authorize',
                               side_effect=[False, True]):
            view_builder = views_types.ViewBuilder()
            now = timeutils.utcnow().isoformat()
            raw_group_type = dict(
                name='new_type',
                description='new_type_desc',
                is_public=True,
                deleted=False,
                created_at=now,
                updated_at=now,
                deleted_at=None,
                id=42,
            )

            request = fakes.HTTPRequest.blank("/v3",
                                              version=mv.GROUP_TYPE)
            output = view_builder.show(request, raw_group_type)

            self.assertIn('group_type', output)
            expected_group_type = dict(
                name='new_type',
                description='new_type_desc',
                is_public=True,
                id=42,
            )
            self.assertDictEqual(expected_group_type, output['group_type'])

    def test_view_builder_show_group_specs_policy(self):
        with mock.patch.object(context.RequestContext,
                               'authorize',
                               side_effect=[True, False]):
            view_builder = views_types.ViewBuilder()
            now = timeutils.utcnow().isoformat()
            raw_group_type = dict(
                name='new_type',
                description='new_type_desc',
                is_public=True,
                deleted=False,
                created_at=now,
                updated_at=now,
                group_specs={},
                deleted_at=None,
                id=42,
            )

            request = fakes.HTTPRequest.blank("/v3",
                                              version=mv.GROUP_TYPE)
            output = view_builder.show(request, raw_group_type)

            self.assertIn('group_type', output)
            expected_group_type = dict(
                name='new_type',
                description='new_type_desc',
                group_specs={},
                is_public=True,
                id=42,
            )
            self.assertDictEqual(expected_group_type, output['group_type'])

    def test_view_builder_show_pass_all_policy(self):
        with mock.patch.object(context.RequestContext,
                               'authorize',
                               side_effect=[True, False]):
            view_builder = views_types.ViewBuilder()
            now = timeutils.utcnow().isoformat()
            raw_group_type = dict(
                name='new_type',
                description='new_type_desc',
                is_public=True,
                deleted=False,
                created_at=now,
                updated_at=now,
                group_specs={},
                deleted_at=None,
                id=42,
            )

            request = fakes.HTTPRequest.blank("/v3",
                                              version=mv.GROUP_TYPE)
            output = view_builder.show(request, raw_group_type)

            self.assertIn('group_type', output)
            expected_group_type = dict(
                name='new_type',
                description='new_type_desc',
                group_specs={},
                is_public=True,
                id=42,
            )
            self.assertDictEqual(expected_group_type, output['group_type'])

    def test_view_builder_list(self):
        view_builder = views_types.ViewBuilder()

        now = timeutils.utcnow().isoformat()
        raw_group_types = []
        for i in range(0, 10):
            raw_group_types.append(
                dict(
                    name='new_type',
                    description='new_type_desc',
                    is_public=True,
                    deleted=False,
                    created_at=now,
                    updated_at=now,
                    group_specs={},
                    deleted_at=None,
                    id=42 + i
                )
            )

        request = fakes.HTTPRequest.blank("/v3",
                                          version=mv.GROUP_TYPE)
        output = view_builder.index(request, raw_group_types)

        self.assertIn('group_types', output)
        for i in range(0, 10):
            expected_group_type = dict(
                name='new_type',
                description='new_type_desc',
                is_public=True,
                id=42 + i
            )
            self.assertDictEqual(expected_group_type,
                                 output['group_types'][i])

    def test_view_builder_list_admin(self):
        view_builder = views_types.ViewBuilder()

        now = timeutils.utcnow().isoformat()
        raw_group_types = []
        for i in range(0, 10):
            raw_group_types.append(
                dict(
                    name='new_type',
                    description='new_type_desc',
                    is_public=True,
                    deleted=False,
                    created_at=now,
                    updated_at=now,
                    group_specs={},
                    deleted_at=None,
                    id=42 + i
                )
            )

        request = fakes.HTTPRequest.blank("/v3", use_admin_context=True,
                                          version=mv.GROUP_TYPE)
        output = view_builder.index(request, raw_group_types)

        self.assertIn('group_types', output)
        for i in range(0, 10):
            expected_group_type = dict(
                name='new_type',
                description='new_type_desc',
                is_public=True,
                group_specs={},
                id=42 + i
            )
            self.assertDictEqual(expected_group_type,
                                 output['group_types'][i])
